// #define VERBOSE

/**
 * @file toy_multi_reg2d.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "fields.h"
#include "yac_interface.h"
#include "geometry.h"
#include "vtk_output.h"
#include "generate_reg2d.h"
#include "test_function.h"

/* ------------------------------------------------- */

/* For simplicity we define the same 8 fields that are in the xml file */

const char * fieldName[] = {"reg2d_conserv_field_out",
                            "unstruct_conserv_field_out",
                            "unstruct_2_conserv_field_out",
                            "cube_conserv_field_out",
                            "reg2d_avg_field_out",
                            "unstruct_avg_field_out",
                            "unstruct_2_avg_field_out",
                            "cube_avg_field_out",
                            "reg2d_hcsbb_field_out",
                            "unstruct_hcsbb_field_out",
                            "unstruct_2_hcsbb_field_out",
                            "cube_hcsbb_field_out",
                            "reg2d_2nd_conserv_field_out",
                            "unstruct_2nd_conserv_field_out",
                            "unstruct_2_2nd_conserv_field_out",
                            "cube_2nd_conserv_field_out",
                            "reg2d_rbf_field_out",
                            "unstruct_rbf_field_out",
                            "unstruct_2_rbf_field_out",
                            "cube_rbf_field_out"};

int main () {

  // Initialisation of MPI

  MPI_Init (0, NULL);

  /* The initialisation phase includes the reading of the XML file */
#ifdef VERBOSE
  printf (". main: calling yac_cinit\n");
#endif

  double tic, toc, time;

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_cinit ("coupling_toy_multi.xml", "coupling.xsd");

  /* The usual component definition, here for two sequential components on the same process */

#ifdef VERBOSE
  printf (". main: calling yac_cdef_comp\n");
#endif

  int comp_id;
  char * comp_name = "ECHAM";

  yac_cdef_comp ( comp_name, &comp_id );
#ifdef VERBOSE
  printf ( ". main: defined %s with local comp ID %i \n", "ECHAM", comp_id );
#endif

  MPI_Comm local_comm;

  yac_cget_localcomm(&local_comm, comp_id);

  int rank, size;
  MPI_Comm_rank(local_comm,&rank);
  MPI_Comm_size(local_comm,&size);
  MPI_Comm_free(&local_comm);

  int ierror = 0;

  int nbr_fields;
  int cell_point_id, corner_point_id;
  
  int field_ids[20];
  int grid_id;

  /* Grid definition for the second component (ECHAM) */

#define NUM_CELLS_X (256)
#define NUM_CELLS_Y (128)

  int total_nbr_cells[2] = {NUM_CELLS_X, NUM_CELLS_Y};
  int num_procs[2];

  generate_reg2d_decomp(total_nbr_cells, size, num_procs);

  int block_pos[2];
  int block_size[2];
  int nbr_cells[2];
  int cyclic[2];
  int nbr_points[2];
  block_pos[0] = rank%num_procs[0];
  block_pos[1] = rank/num_procs[0];
  block_size[0] = (NUM_CELLS_X + num_procs[0] - 1)/num_procs[0];
  block_size[1] = (NUM_CELLS_Y + num_procs[1] - 1)/num_procs[1];
  nbr_cells[0] = MIN(block_size[0], NUM_CELLS_X - block_size[0] * block_pos[0]);
  nbr_cells[1] = MIN(block_size[1], NUM_CELLS_Y - block_size[1] * block_pos[1]);
  cyclic[0] = num_procs[0] == 1;
  cyclic[1] = 0;

  // halo
  if (!cyclic[0])
   nbr_cells[0] += 2;
  if (num_procs[1] > 1) {

    nbr_cells[1] += 1;

    if ((block_pos[1] != 0) && (block_pos[1] != num_procs[1]-1))
      nbr_cells[1] += 1;
  }

  if (cyclic[0])
    nbr_points[0] = nbr_cells[0];
  else
    nbr_points[0] = nbr_cells[0]+1;

  nbr_points[1] = nbr_cells[1] + 1;

  double * global_x_vertices = xmalloc((NUM_CELLS_X+3) * sizeof(*global_x_vertices));
  double * global_y_vertices = xmalloc((NUM_CELLS_Y+1) * sizeof(*global_y_vertices));;
  double * x_vertices;
  double * y_vertices;
  double * x_corner_points;
  double * y_corner_points;

  for (int i = 0; i < NUM_CELLS_X+3; ++i)
    global_x_vertices[i] =
      0.0 + (2.0 * M_PI / ((double)NUM_CELLS_X)) * (double)(i-1);
  for (int i = 0; i < NUM_CELLS_Y; ++i)
    global_y_vertices[i] =
      -M_PI_2 + (M_PI / ((double)NUM_CELLS_Y)) * (double)i;
  global_y_vertices[NUM_CELLS_Y] = M_PI_2;

  if (cyclic[0])
    x_vertices = x_corner_points = global_x_vertices + 1;
  else
    x_vertices = x_corner_points = global_x_vertices + block_size[0] * block_pos[0];

  y_vertices = y_corner_points = global_y_vertices + block_size[1] * block_pos[1];

  if (block_pos[1] != 0) {
    y_vertices -= 1;
    y_corner_points -= 1;
  }

  double * x_cell_points = xmalloc(nbr_cells[0] * sizeof(*x_cell_points));
  double * y_cell_points = xmalloc(nbr_cells[1] * sizeof(*y_cell_points));

  for (int i = 0; i < nbr_cells[0]; ++i)
    x_cell_points[i] = (x_corner_points[i] + x_corner_points[i + 1]) * 0.5;
  for (int i = 0; i < nbr_cells[1]; ++i)
    y_cell_points[i] = (y_corner_points[i] + y_corner_points[i + 1]) * 0.5;

  yac_cdef_grid_reg2d ( "echam_grid", nbr_points, cyclic, x_vertices, y_vertices, &grid_id);

  int * global_global_cell_id = xmalloc(NUM_CELLS_Y * (NUM_CELLS_X + 2) * sizeof(*global_global_cell_id));
  int * global_global_corner_id = xmalloc((NUM_CELLS_Y + 1) * (NUM_CELLS_X + 3) * sizeof(*global_global_corner_id));

  for (int j = 0, id = 0; j < NUM_CELLS_Y; ++j)
    for (int i = 1; i <= NUM_CELLS_X; ++i)
      global_global_cell_id[j * (NUM_CELLS_X + 2) + i] = id++;
  for (int i = 0; i < NUM_CELLS_Y; ++i) {
    global_global_cell_id[i * (NUM_CELLS_X + 2) + 0] = global_global_cell_id[i * (NUM_CELLS_X + 2) + NUM_CELLS_X];
    global_global_cell_id[i * (NUM_CELLS_X + 2) + NUM_CELLS_X+1] = global_global_cell_id[i * (NUM_CELLS_X + 2) + 1];
  }

  if (num_procs[0] == 1) {

    for (int j = 0, id = 0; j < NUM_CELLS_Y + 1; ++j)
      for (int i = 1; i <= NUM_CELLS_X; ++i)
        global_global_corner_id[j * (NUM_CELLS_X + 3) + i] = id++;

  } else {

    for (int j = 0, id = 0; j < NUM_CELLS_Y + 1; ++j)
      for (int i = 0; i < NUM_CELLS_X; ++i)
        global_global_corner_id[j * (NUM_CELLS_X + 3) + (i + 1)] = id++;

    for (int i = 0; i < NUM_CELLS_Y + 1; ++i) {

      global_global_corner_id[i * (NUM_CELLS_X + 3) + 0] =
        global_global_corner_id[i * (NUM_CELLS_X + 3) + NUM_CELLS_X];
      global_global_corner_id[i * (NUM_CELLS_X + 3) + NUM_CELLS_X + 1] =
        global_global_corner_id[i * (NUM_CELLS_X + 3) + 1];
      global_global_corner_id[i * (NUM_CELLS_X + 3) + NUM_CELLS_X + 2] =
        global_global_corner_id[i * (NUM_CELLS_X + 3) + 2];
    }
  }

  int * local_global_cell_id = xmalloc(nbr_cells[1] * nbr_cells[0] * sizeof(*local_global_cell_id));
  int * cell_core_mask = xmalloc(nbr_cells[1] * nbr_cells[0] * sizeof(*cell_core_mask));
  int * local_global_corner_id = xmalloc(nbr_points[1] * nbr_points[0] * sizeof(*local_global_corner_id));
  int * corner_core_mask = xmalloc(nbr_points[1] * nbr_points[0] * sizeof(*corner_core_mask));

  int offset_x, offset_y;

  if (cyclic[0])
    offset_x = 1;
  else
    offset_x = block_size[0] * block_pos[0];
  offset_y = block_size[1] * block_pos[1] - (block_pos[1] != 0);

  for (int j = 0; j < nbr_cells[1]; ++j)
    for (int i = 0; i < nbr_cells[0]; ++i)
      local_global_cell_id[j * nbr_cells[0] + i] =
        global_global_cell_id[(j+offset_y) * (NUM_CELLS_X + 2) + i+offset_x];

  free(global_global_cell_id);

  for (int j = 0; j < nbr_points[1]; ++j)
    for (int i = 0; i < nbr_points[0]; ++i)
      local_global_corner_id[j * nbr_points[0] + i] =
        global_global_corner_id[(j+offset_y) * (NUM_CELLS_X + 3) +i+offset_x];

  free(global_global_corner_id);

  for (int j = 0; j < nbr_cells[1]; ++j)
    for (int i = 0; i < nbr_cells[0]; ++i)
      cell_core_mask[j * nbr_cells[0] + i] = 1;
  if (num_procs[1] > 1) {
    if (block_pos[1] > 0)
      for (int i = 0; i < nbr_cells[0]; ++i)
        cell_core_mask[i] = 0;
    if (block_pos[1] < num_procs[1] - 1)
      for (int i = 0; i < nbr_cells[0]; ++i)
        cell_core_mask[(nbr_cells[1] - 1) * nbr_cells[0] + i] = 0;
  }
  if (num_procs[0] > 1) {
    for (int j = 0; j < nbr_cells[1]; ++j) {
      cell_core_mask[j * nbr_cells[0]] = 0;
      cell_core_mask[j * nbr_cells[0] + nbr_cells[0] - 1] = 0;
    }
  }

  for (int j = 0; j < nbr_points[1]; ++j)
    for (int i = 0; i < nbr_points[0]; ++i)
      corner_core_mask[j * nbr_points[0] + i] = 1;
  if (num_procs[1] > 1) {
    if (block_pos[1] > 0)
      for (int i = 0; i < nbr_points[0]; ++i)
        corner_core_mask[i] = 0;
    if (block_pos[1] < num_procs[1] - 1)
      for (int i = 0; i < nbr_points[0]; ++i)
        corner_core_mask[(nbr_points[1] - 1) * nbr_points[0] + i] = 0;
  }
  if (num_procs[0] > 1) {
    for (int j = 0; j < nbr_points[1]; ++j) {
      corner_core_mask[j * nbr_points[0]] = 0;
      corner_core_mask[j * nbr_points[0] + nbr_points[0] - 1] = 0;
    }
  }

  yac_cset_global_index(local_global_cell_id, CELL, grid_id);
  yac_cset_core_mask(cell_core_mask, CELL, grid_id);
  yac_cset_global_index(local_global_corner_id, CORNER, grid_id);
  yac_cset_core_mask(corner_core_mask, CORNER, grid_id);

  yac_cdef_points_reg2d ( grid_id, nbr_cells, CELL, x_cell_points, y_cell_points, &cell_point_id );
  yac_cdef_points_reg2d ( grid_id, nbr_points, CORNER, x_corner_points, y_corner_points, &corner_point_id );

  /* Field definition for the second component (ECHAM) */

  nbr_fields = 0;
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+0], comp_id, &cell_point_id, 1, &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+4], comp_id, &corner_point_id, 1, &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+8], comp_id, &cell_point_id, 1, &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+12], comp_id, &cell_point_id, 1, &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(fieldName[i+16], comp_id, &cell_point_id, 1, &field_ids[nbr_fields++]);

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_reg2d: Time for initialisation %f\n", time ); 

  /* Search. */

#ifdef VERBOSE
  printf (". main: calling yac_csearch\n");
#endif

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_csearch ( &ierror );

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_reg2d: Time for search %f\n", time ); 

  unsigned num_cells = nbr_cells[0] * nbr_cells[1];
  unsigned num_corners = nbr_points[0] * nbr_points[1];

  double * conserv_in_unstruct_point_data = xmalloc(num_cells * sizeof(*conserv_in_unstruct_point_data));
  double * conserv_in_unstruct_2_point_data = xmalloc(num_cells * sizeof(*conserv_in_unstruct_2_point_data));
  double * conserv_in_cube_point_data = xmalloc(num_cells * sizeof(*conserv_in_cube_point_data));

  double * avg_in_unstruct_point_data = xmalloc(num_corners * sizeof(*avg_in_unstruct_point_data));
  double * avg_in_unstruct_2_point_data = xmalloc(num_corners * sizeof(*avg_in_unstruct_2_point_data));
  double * avg_in_cube_point_data = xmalloc(num_corners * sizeof(*avg_in_cube_point_data));

  double * hcsbb_in_unstruct_point_data = xmalloc(num_cells * sizeof(*hcsbb_in_unstruct_point_data));
  double * hcsbb_in_unstruct_2_point_data = xmalloc(num_cells * sizeof(*hcsbb_in_unstruct_2_point_data));
  double * hcsbb_in_cube_point_data = xmalloc(num_cells * sizeof(*hcsbb_in_cube_point_data));

  double * conserv_2nd_in_unstruct_point_data = xmalloc(num_cells * sizeof(*conserv_2nd_in_unstruct_point_data));
  double * conserv_2nd_in_unstruct_2_point_data = xmalloc(num_cells * sizeof(*conserv_2nd_in_unstruct_2_point_data));
  double * conserv_2nd_in_cube_point_data = xmalloc(num_cells * sizeof(*conserv_2nd_in_cube_point_data));

  double * rbf_in_unstruct_point_data = xmalloc(num_cells * sizeof(*rbf_in_unstruct_point_data));
  double * rbf_in_unstruct_2_point_data = xmalloc(num_cells * sizeof(*rbf_in_unstruct_2_point_data));
  double * rbf_in_cube_point_data = xmalloc(num_cells * sizeof(*rbf_in_cube_point_data));

  for (unsigned i = 0; i < num_cells; ++i) {
    conserv_in_unstruct_point_data[i] = -10;
    conserv_in_unstruct_2_point_data[i] = -10;
    conserv_in_cube_point_data[i] = -10;

    hcsbb_in_unstruct_point_data[i] = -10;
    hcsbb_in_unstruct_2_point_data[i] = -10;
    hcsbb_in_cube_point_data[i] = -10;

    conserv_2nd_in_unstruct_point_data[i] = -10;
    conserv_2nd_in_unstruct_2_point_data[i] = -10;
    conserv_2nd_in_cube_point_data[i] = -10;

    rbf_in_unstruct_point_data[i] = -10;
    rbf_in_unstruct_2_point_data[i] = -10;
    rbf_in_cube_point_data[i] = -10;
  }

  for (unsigned i = 0; i < num_corners; ++i) {
    avg_in_unstruct_point_data[i] = -10;
    avg_in_unstruct_2_point_data[i] = -10;
    avg_in_cube_point_data[i] = -10;
  }

  double * cell_out_point_data = xmalloc(num_cells * sizeof(*cell_out_point_data));
  double * corner_out_point_data = xmalloc(num_corners * sizeof(*corner_out_point_data));

  int err;
  int info;

  unsigned * cell_corners = xmalloc(num_cells * 4 * sizeof(*cell_corners));
  unsigned * num_points_per_cell = xmalloc(num_cells * sizeof(*num_points_per_cell));
  for (unsigned i = 0; i < num_cells; ++i) {

    {
      unsigned x_index, y_index;

      y_index = i / nbr_cells[0];
      x_index = i - y_index * nbr_cells[0];

      if (!cyclic[0]) {
      
        cell_corners[i*4+0] =  y_index      * (nbr_cells[0] + 1) + x_index;
        cell_corners[i*4+1] =  y_index      * (nbr_cells[0] + 1) + x_index + 1;
        cell_corners[i*4+2] = (y_index + 1) * (nbr_cells[0] + 1) + x_index + 1;
        cell_corners[i*4+3] = (y_index + 1) * (nbr_cells[0] + 1) + x_index;

      } else {
      
        cell_corners[i*4+0] = y_index * nbr_cells[0] + x_index;
        if (x_index + 1 != nbr_cells[0]) {
          cell_corners[i*4+1] =  y_index      * nbr_cells[0] + x_index + 1;
          cell_corners[i*4+2] = (y_index + 1) * nbr_cells[0] + x_index + 1;
        } else {
          cell_corners[i*4+1] =  y_index      * nbr_cells[0];
          cell_corners[i*4+2] = (y_index + 1) * nbr_cells[0];
        }
        cell_corners[i*4+3] = (y_index + 1) * nbr_cells[0] + x_index;
      }
    }
    num_points_per_cell[i] = 4;
  }

  for (unsigned i = 0; i < num_cells; ++i) {

    double curr_x, curr_y;

    curr_x = (x_vertices[cell_corners[i*4+0]%nbr_points[0]] +
              x_vertices[cell_corners[i*4+1]%nbr_points[0]]) * 0.5;
    curr_y = (y_vertices[cell_corners[i*4+1]/nbr_points[0]] +
              y_vertices[cell_corners[i*4+2]/nbr_points[0]]) * 0.5;

    cell_out_point_data[i] = test_func(curr_x, curr_y);
  }

  for (unsigned i = 0; i < num_corners; ++i)
    corner_out_point_data[i] = test_func(x_corner_points[i%nbr_points[0]],
                                         y_corner_points[i/nbr_points[0]]);

  MPI_Barrier(MPI_COMM_WORLD);
  
  tic=MPI_Wtime();

  {
    double *point_set_data[1];
    double **collection_data[1] = {point_set_data};

    point_set_data[0] = cell_out_point_data;
    yac_cput(field_ids[0], 1, collection_data, &info, &err);
    point_set_data[0] = corner_out_point_data;
    yac_cput(field_ids[4], 1, collection_data, &info, &err);
    point_set_data[0] = cell_out_point_data;
    yac_cput(field_ids[8], 1, collection_data, &info, &err);
    point_set_data[0] = cell_out_point_data;
    yac_cput(field_ids[12], 1, collection_data, &info, &err);
    point_set_data[0] = cell_out_point_data;
    yac_cput(field_ids[16], 1, collection_data, &info, &err);
  }

  {
    double *collection_data[1];

    collection_data[0] = conserv_in_unstruct_point_data;
    yac_cget(field_ids[1], 1, collection_data, &info, &err);
    collection_data[0] = conserv_in_unstruct_2_point_data;
    yac_cget(field_ids[2], 1, collection_data, &info, &err);
    collection_data[0] = conserv_in_cube_point_data;
    yac_cget(field_ids[3], 1, collection_data, &info, &err);

    collection_data[0] = avg_in_unstruct_point_data;
    yac_cget(field_ids[5], 1, collection_data, &info, &err);
    collection_data[0] = avg_in_unstruct_2_point_data;
    yac_cget(field_ids[6], 1, collection_data, &info, &err);
    collection_data[0] = avg_in_cube_point_data;
    yac_cget(field_ids[7], 1, collection_data, &info, &err);

    collection_data[0] = hcsbb_in_unstruct_point_data;
    yac_cget(field_ids[9], 1, collection_data, &info, &err);
    collection_data[0] = hcsbb_in_unstruct_2_point_data;
    yac_cget(field_ids[10], 1, collection_data, &info, &err);
    collection_data[0] = hcsbb_in_cube_point_data;
    yac_cget(field_ids[11], 1, collection_data, &info, &err);

    collection_data[0] = conserv_2nd_in_unstruct_point_data;
    yac_cget(field_ids[13], 1, collection_data, &info, &err);
    collection_data[0] = conserv_2nd_in_unstruct_2_point_data;
    yac_cget(field_ids[14], 1, collection_data, &info, &err);
    collection_data[0] = conserv_2nd_in_cube_point_data;
    yac_cget(field_ids[15], 1, collection_data, &info, &err);

    collection_data[0] = rbf_in_unstruct_point_data;
    yac_cget(field_ids[17], 1, collection_data, &info, &err);
    collection_data[0] = rbf_in_unstruct_2_point_data;
    yac_cget(field_ids[18], 1, collection_data, &info, &err);
    collection_data[0] = rbf_in_cube_point_data;
    yac_cget(field_ids[19], 1, collection_data, &info, &err);
  }

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_reg2d: Time for ping-pong %f\n", time );

  //----------------------------------------------------------
  // write field to vtk output file
  //----------------------------------------------------------

  double * point_data = xmalloc(nbr_points[0] * nbr_points[1] * 3 * sizeof(*point_data));
  for (int i = 0; i < nbr_points[1]; ++i)
    for (int j = 0; j < nbr_points[0]; ++j)
      LLtoXYZ(
        x_vertices[j], y_vertices[i], &point_data[3*(i * nbr_points[0] + j)]);

  char vtk_filename[32];

  sprintf(vtk_filename, "reg2d_out_%d.vtk", rank);

  VTK_FILE *vtk_file = vtk_open(vtk_filename, "reg2d_out");
  vtk_write_point_data(vtk_file, point_data, nbr_points[0]*nbr_points[1]);
  vtk_write_cell_data(vtk_file, cell_corners, num_points_per_cell, num_cells);
  vtk_write_point_scalars_int(vtk_file, corner_core_mask,
                              nbr_points[0]*nbr_points[1], "corner_core_mask");
  vtk_write_point_scalars_int(vtk_file, local_global_corner_id,
                              nbr_points[0]*nbr_points[1], "global_corner_id");
  vtk_write_cell_scalars_int(vtk_file, cell_core_mask, num_cells,
                             "cell_core_mask");
  vtk_write_cell_scalars_int(vtk_file, local_global_cell_id, num_cells,
                             "global_cell_id");

  vtk_write_cell_scalars_double(vtk_file, conserv_in_unstruct_point_data, num_cells,
                                "conserv_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, conserv_in_unstruct_2_point_data, num_cells,
                                "conserv_in_unstruct_2");
  vtk_write_cell_scalars_double(vtk_file, conserv_in_cube_point_data, num_cells,
                                "conserv_in_cube");

  vtk_write_point_scalars_double(vtk_file, avg_in_unstruct_point_data, num_corners,
                                 "avg_in_unstruct");
  vtk_write_point_scalars_double(vtk_file, avg_in_unstruct_2_point_data, num_corners,
                                 "avg_in_unstruct_2");
  vtk_write_point_scalars_double(vtk_file, avg_in_cube_point_data, num_corners,
                                 "avg_in_cube");

  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_unstruct_point_data, num_cells,
                                "hcsbb_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_unstruct_2_point_data, num_cells,
                                "hcsbb_in_unstruct_2");
  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_cube_point_data, num_cells,
                                "hcsbb_in_cube");

  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_unstruct_point_data, num_cells,
                                "2nd_conserv_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_unstruct_2_point_data, num_cells,
                                "2nd_conserv_in_unstruct_2");
  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_cube_point_data, num_cells,
                                "2nd_conserv_in_cube");

  vtk_write_cell_scalars_double(vtk_file, rbf_in_unstruct_point_data, num_cells,
                                "rbf_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, rbf_in_unstruct_2_point_data, num_cells,
                                "rbf_in_unstruct_2");
  vtk_write_cell_scalars_double(vtk_file, rbf_in_cube_point_data, num_cells,
                                "rbf_in_cube");

  vtk_write_cell_scalars_double(vtk_file, cell_out_point_data, num_cells, "cell_out");
  vtk_write_point_scalars_double(vtk_file, corner_out_point_data, num_corners, "corner_out");

  vtk_close(vtk_file);

  yac_cfinalize();

  MPI_Finalize();

  free(num_points_per_cell);
  free(cell_corners);
  free(point_data);
  free(corner_out_point_data);
  free(cell_out_point_data);
  free(conserv_in_unstruct_point_data);
  free(conserv_in_unstruct_2_point_data);
  free(conserv_in_cube_point_data);
  free(avg_in_unstruct_point_data);
  free(avg_in_unstruct_2_point_data);
  free(avg_in_cube_point_data);
  free(hcsbb_in_unstruct_point_data);
  free(hcsbb_in_unstruct_2_point_data);
  free(hcsbb_in_cube_point_data);
  free(conserv_2nd_in_unstruct_point_data);
  free(conserv_2nd_in_unstruct_2_point_data);
  free(conserv_2nd_in_cube_point_data);
  free(rbf_in_unstruct_point_data);
  free(rbf_in_unstruct_2_point_data);
  free(rbf_in_cube_point_data);
  free(corner_core_mask);
  free(local_global_corner_id);
  free(cell_core_mask);
  free(local_global_cell_id);
  free(x_cell_points);
  free(y_cell_points);
  free(global_x_vertices);
  free(global_y_vertices);

  return EXIT_SUCCESS;
}

