// #define VERBOSE

/**
 * @file toy_scrip.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "yac_config.h"

#if defined YAC_NETCDF_ENABLED
#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <netcdf.h>
#include <string.h>
#include "utils.h"
#include "fields.h"
#include "yac_interface.h"
#include "geometry.h"
#include "vtk_output.h"
#include "read_scrip_grid.h"
#include "test_function.h"
#include "io_utils.h"

static void write_config_xml_file(
  char const * filename,
  size_t num_components, char const * component_names[],
  size_t num_interpolations, char const * interpolation_names[],
  char const * interpolation_xml_string[],
  bool const write_weights);
static int * read_mask(char const * filename, char const * grid_name, int const valid_mask_value);

int main (int argc, char *argv[]) {

  /* Choose the experiment between
     "gen_mask" : all the grids, only interpolate the constant one
                  function with conservative 1st ord destarea
     "noicoh" : torc, nogt, bggd, sse7, icos grids, all the interpolations
     "nogt_icoh" : nogt, icoh grids, all the interpolations
     "icos_icoh" : icos, icoh grids, all the interpolations
  */
  char *experiment = "noicoh";

  // Initialisation of MPI

  MPI_Init (0, NULL);

  bool write_weights;
  size_t num_components;
  size_t num_interpolations;
  char const ** component_names;

  char const * component_names_full[] = {
    "torc"
    ,"bggd"
    ,"nogt"
    ,"sse7"
    ,"icos"
    ,"icoh"
    };
  char const * component_names_nogt_icoh[] = {
    "nogt"
    ,"icoh"
    };
  char const * component_names_icos_icoh[] = {
    "icos"
    ,"icoh"
    };
  char const * interpolation_names[] = {
    "CONSERV_DESTAREA"
    ,"CONSERV_FRACAREA"
    ,"CONS2ND_FRACAREA"
    ,"DISTWGT_4"
    ,"DISTWGT_1"
    ,"HCSBB"
    ,"AVG_DIST"
    };
  char const * interpolation_xml_string[] = {
    "<interpolation method=\"conservative\" "
                   "enforced_conservation=\"true\" "
                   "normalisation=\"DESTAREA\" "
                   "partial_coverage=\"true\" />\n"
    "<interpolation method=\"fixed_value\" user_value=\"0.0\"/>\n",

    "<interpolation method=\"conservative\" "
                   "normalisation=\"FRACAREA\" "
                   "partial_coverage=\"true\" />\n"
    "<interpolation method=\"fixed_value\" user_value=\"0.0\"/>\n",

    "<interpolation method=\"conservative\" "
                   "normalisation=\"FRACAREA\" "
                   "partial_coverage=\"true\" "
                   "order=\"2\" />\n"
    "<interpolation method=\"fixed_value\" user_value=\"0.0\"/>\n",

    "<interpolation method=\"n-nearest_neighbor\" "
                   "n=\"4\" weighted=\"DISTANCE_WEIGHTED\"/>",

    "<interpolation method=\"n-nearest_neighbor\" "
                   "n=\"1\" weighted=\"ARITHMETIC_AVERAGE\"/>",

    "<interpolation method=\"bernstein_bezier\" />\n"
    "<interpolation method=\"n-nearest_neighbor\" "
                   "n=\"4\" weighted=\"DISTANCE_WEIGHTED\"/>\n"
    "<interpolation method=\"fixed_value\" user_value=\"0.0\"/>\n",

    "<interpolation method=\"average\" "
                   "weighted=\"DISTANCE_WEIGHTED\" "
                   "partial_coverage=\"true\"/>\n"
    "<interpolation method=\"n-nearest_neighbor\" "
                   "n=\"2\" weighted=\"DISTANCE_WEIGHTED\"/>\n"
  };

  if (strcmp(experiment,"gen_mask") == 0) {
    write_weights = 0;
    num_components = 6;
    num_interpolations = 1;
    component_names = component_names_full;
  } else if (strcmp(experiment,"noicoh") == 0) {
    write_weights = 1;
    num_components = 5;
    num_interpolations = 7;
    component_names = component_names_full;
  } else if (strcmp(experiment,"nogt_icoh") == 0) {
    write_weights = 1;
    num_components = 2;
    num_interpolations = 7;
    component_names = component_names_nogt_icoh;
  } else if (strcmp(experiment,"icos_icoh") == 0) {
    write_weights = 1;
    num_components = 2;
    num_interpolations = 7;
    component_names = component_names_icos_icoh;
  }

  char * xml_config_filename = "coupling_scrip.xml";

  int rank, size;
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  if ((size_t)size != num_components) {
    fprintf(
      stderr, "Wrong number of processes (has to be %zu)\n", num_components);
    exit(EXIT_FAILURE);
  }

  // generate config xml file
  if (rank == 0)
    write_config_xml_file(xml_config_filename,
                          num_components, component_names,
                          num_interpolations, interpolation_names,
                          interpolation_xml_string, write_weights);
  MPI_Barrier(MPI_COMM_WORLD);

  yac_cinit (xml_config_filename, "coupling.xsd");

  // register component
  int comp_id;
  char const * comp_name = component_names[rank];
  yac_cdef_comp(comp_name, &comp_id);

  // read grid information
  char * grid_filename = "OASIS-grid/grids.nc";
  char * mask_filename = "OASIS-grid/masks_no_atm.nc";
  int valid_mask_value = 0;
  size_t num_vertices;
  size_t num_cells;
  int * num_vertices_per_cell;
  double * x_vertices;
  double * y_vertices;
  double * x_cell;
  double * y_cell;
  int * cell_to_vertex;
  int * cell_core_mask;
  read_scrip_grid_information(
    grid_filename, mask_filename, comp_name, valid_mask_value,
    &num_vertices, &num_cells, &num_vertices_per_cell,
    &x_vertices, &y_vertices, &x_cell, &y_cell, &cell_to_vertex,
    &cell_core_mask);

  // register grid
  int grid_id;
  yac_cdef_grid_unstruct(
    comp_name, (int)num_vertices, (int)num_cells, num_vertices_per_cell,
    x_vertices, y_vertices, cell_to_vertex, &grid_id);

  // set global ids and core masks
  yac_int * global_cell_ids = xmalloc(num_cells * sizeof(*global_cell_ids));
  yac_int * global_vertex_ids =
    xmalloc(num_vertices * sizeof(*global_vertex_ids));
  for (size_t i = 0; i < num_cells; ++i) global_cell_ids[i] = i;
  for (size_t i = 0; i < num_vertices; ++i) global_vertex_ids[i] = i;
  yac_cset_global_index(global_cell_ids, CELL, grid_id);
  yac_cset_core_mask(cell_core_mask, CELL, grid_id);
  yac_cset_global_index(global_vertex_ids, CORNER, grid_id);

  int * nogt_mask = read_mask("masks_nogt_yac.nc", comp_name, 0);
  int * torc_mask = read_mask("masks_torc_yac.nc", comp_name, 0);

  // register cell middle points and mask
  int cell_point_id_no_mask,
      cell_point_id_nogt_mask,
      cell_point_id_torc_mask;
  yac_cdef_points_unstruct(
    grid_id, (int)num_cells, CELL, x_cell, y_cell, &cell_point_id_no_mask);
  if (nogt_mask != NULL) {
    yac_cdef_points_unstruct(
      grid_id, (int)num_cells, CELL, x_cell, y_cell, &cell_point_id_nogt_mask);
    yac_cset_mask(nogt_mask, cell_point_id_nogt_mask);
  } else cell_point_id_nogt_mask = cell_point_id_no_mask;
  if (torc_mask != NULL) {
    yac_cdef_points_unstruct(
      grid_id, (int)num_cells, CELL, x_cell, y_cell, &cell_point_id_torc_mask);
    yac_cset_mask(torc_mask, cell_point_id_torc_mask);
  } else cell_point_id_torc_mask = cell_point_id_no_mask;

  // register in fields
  int field_ids[num_components][num_interpolations];
  for (size_t i = 0; i < num_components; ++i) {
    int cell_point_id = cell_point_id_no_mask;
    if (!strcmp("nogt", component_names[i]))
      cell_point_id = cell_point_id_nogt_mask;
    if (!strcmp("torc", component_names[i]))
      cell_point_id = cell_point_id_torc_mask;
    for (size_t j = 0; j < num_interpolations; ++j) {
      char field_name[1024];
      snprintf(
        field_name, 1024, "%s_%s_out",
        component_names[i], interpolation_names[j]);
      yac_cdef_field(
        field_name, comp_id, &cell_point_id, 1, &field_ids[i][j]);
    }
  }

  // initialise coupling
  int ierror;
  yac_csearch(&ierror);

  // initialise vtk output file
  VTK_FILE *vtk_file;
  {
    char vtk_filename[1024];
    sprintf(vtk_filename, "%s.vtk", comp_name);

    double (*point_data)[3] = xmalloc(num_vertices * sizeof(*point_data));
    for (size_t i = 0; i < num_vertices; ++i)
      LLtoXYZ(x_vertices[i], y_vertices[i], point_data[i]);

    vtk_file = vtk_open(vtk_filename, comp_name);
    vtk_write_point_data(vtk_file, &point_data[0][0], (int)num_vertices);
    vtk_write_cell_data(
      vtk_file, (unsigned *)cell_to_vertex,
      (unsigned*)num_vertices_per_cell, (int)num_cells);
    vtk_write_point_scalars_int(
      vtk_file, global_vertex_ids, (int)num_vertices, "global_vertex_id");
    vtk_write_cell_scalars_int(
      vtk_file, cell_core_mask, (int)num_cells, "cell_core_mask");
    vtk_write_cell_scalars_int(
      vtk_file, global_cell_ids, (int)num_cells, "global_cell_id");
    if (nogt_mask != NULL)
      vtk_write_cell_scalars_int(
        vtk_file, nogt_mask, (int)num_cells, "nogt_mask");
    if (torc_mask != NULL)
      vtk_write_cell_scalars_int(
        vtk_file, torc_mask, (int)num_cells, "torc_mask");
    free(point_data);
  }

  if (nogt_mask) free(nogt_mask);
  if (torc_mask) free(torc_mask);

  // allocate memory for field data
  double * out_data = xmalloc(num_cells * sizeof(*out_data));
  double * in_data = xmalloc(num_cells * sizeof(*in_data));

  // array of test functions
  double (*test_functions[])(double lon, double lat) = {
    test_func,
    test_ana_fcos,
    test_ana_fcossin,
    test_one,
    test_gulfstream,
    test_harmonic,
    test_vortex
  };
  char * test_func_names[] = {
    "yac",
    "fcos",
    "fcossin",
    "one",
    "gulfstream",
    "harmonic",
    "vortex"
  };
  size_t num_test_function = 7;

  // "time loop"
  for (size_t i = 0; i < num_test_function; ++i) {

    int info, err;

    // generate out data
    for (size_t j = 0; j < num_cells; ++j)
      out_data[j] = test_functions[i](x_cell[j], y_cell[j]);

    // write out field to vtk file
    {
      char field_name[1024];
      snprintf(field_name, 1024, "test_%s_out", test_func_names[i]);
      vtk_write_cell_scalars_double(
        vtk_file, out_data, (int)num_cells, field_name);
    }

    // put out fields
    double *point_set_data[1] = {out_data};
    double **collection_data[1] = {point_set_data};
    for (size_t j = 0; j < num_interpolations; ++j)
      yac_cput(field_ids[rank][j], 1, collection_data, &info, &err);

    // get in fields
    for (size_t j = 0; j < num_components; ++j) {
      if ((int)j == rank) continue;
      for (size_t k = 0; k < num_interpolations; ++k) {
        for (size_t l = 0; l < num_cells; ++l) in_data[l] = 0.0; //-10.0;
        double *collection_data[1] = {in_data};
        yac_cget(field_ids[j][k], 1, collection_data, &info, &err);

        // write in field to vtk file
        {
          char field_name[1024];
          snprintf(
            field_name, 1024, "test_%s_%s_%s",
            test_func_names[i], component_names[j], interpolation_names[k]);
          vtk_write_cell_scalars_double(
            vtk_file, in_data, (int)num_cells, field_name);
        }
      }
    }
  }

  // close vtk file
  {
    vtk_close(vtk_file);
  }

  // finalize YAC and MPI
  yac_cfinalize();
  MPI_Finalize();

  free(in_data);
  free(out_data);
  free(num_vertices_per_cell);
  free(x_vertices);
  free(y_vertices);
  free(x_cell);
  free(y_cell);
  free(cell_to_vertex);
  free(cell_core_mask);
  free(global_vertex_ids);
  free(global_cell_ids);

  return EXIT_SUCCESS;
}

static void write_config_xml_file(
  char const * filename,
  size_t num_components, char const * component_names[],
  size_t num_interpolations, char const * interpolation_names[],
  char const * interpolation_xml_string[],
  bool const write_weights) {

  FILE * xmlfile = fopen(filename, "w+");

  // write header
  fputs(
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\n"
    "<!-- Example XML document -->\n"
    "<coupling\n"
    "	xmlns=\"http://www.w3schools.com\"\n"
    "	xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n"
    "	xsi:schemaLocation=\"http://www.w3schools.com coupling.xsd\">\n\n"
    "	<redirect redirect_of_root=\"false\" redirect_stdout=\"false\"/>\n\n",
    xmlfile);

  // register components
  fputs("	<components>\n", xmlfile);
  for (size_t i = 0; i < num_components; ++i) {
    fprintf(xmlfile,
      "		<component id=\"%zu\">\n"
      "			<name>%s</name>\n"
      "			<model>dummy</model>\n"
      "			<simulated>dummy</simulated>\n"
      "			<transient_grid_refs>\n",
      i+1, component_names[i]);
    // mention all fields that will be coupled with this component
    // (all fields are involved with all components)
    for (size_t j = 0, l = 1; j < num_components; ++j)
      for (size_t k = 0; k < num_interpolations; ++k, ++l)
        fprintf(xmlfile,
          "				<transient_grid_ref id=\"%zu\"  "
                  "transient_ref=\"%zu\" "
                  "grid_ref=\"%zu\" "
                  "collection_size=\"1\"/>\n", l, l, i+1);
    fprintf(xmlfile,
      "			</transient_grid_refs>\n"
      "		</component>\n");
  }
  fputs("	</components>\n\n", xmlfile);

  // register fields
  fputs("	<transients>\n", xmlfile);
  for (size_t i = 0, l = 1; i < num_components; ++i)
    for (size_t j = 0; j < num_interpolations; ++j, ++l)
      fprintf(xmlfile,
        "		<transient id=\"%zu\"  "
            "transient_standard_name=\"%s_%s_out\" />\n",
        l, component_names[i], interpolation_names[j]);
  fputs("	</transients>\n\n", xmlfile);

  // register grids
  fputs("	<grids>\n", xmlfile);
  for (size_t i = 0; i < num_components; ++i)
    fprintf(xmlfile,
      "		<grid id=\"%zu\"  alias_name=\"%s\" />\n", i+1, component_names[i]);
  fputs("	</grids>\n\n", xmlfile);

  // time stuff
  fputs("	<dates>\n"
        "		<start_date>2008-03-09T16:05:07</start_date>\n"
        "		<end_date>2008-03-10T16:05:07</end_date>\n"
        "		<calendar>proleptic-gregorian</calendar>\n"
        "	</dates>\n"
        "	<timestep_unit>second</timestep_unit>\n\n", xmlfile);

  // register coupling between components
  fputs("	<couples>\n", xmlfile);
  // couple all components with each other
  for (size_t i = 0; i < num_components; ++i) {
    for (size_t j = i + 1; j < num_components; ++j) {
      fprintf(xmlfile,
        "		<couple>\n"
        "			<component1 component_id=\"%zu\" />\n"
        "			<component2 component_id=\"%zu\" />\n",
        i+1, j+1);
      // for all interpolation methods
      for (size_t k = 0; k < num_interpolations; ++k) {
        // for both directions
        size_t src_comp_id = i;
        size_t tgt_comp_id = j;
        for (int l = 1; l <= 2; ++l) {
          char weight_filename[1024];
          snprintf(
            weight_filename, 1024, "rmp_%s_to_%s_yac_%s.nc",
            component_names[src_comp_id], component_names[tgt_comp_id],
            interpolation_names[k]);
          fprintf(xmlfile,
            "			<transient_couple transient_id=\"%zu\">\n"
            "				<source component_ref=\"%d\" "
                           "transient_grid_ref=\"%zu\" />\n"
            "				<target transient_grid_ref=\"%zu\" />\n"
            "				<timestep>\n"
            "					<source>10</source>\n"
            "					<target>10</target>\n"
            "					<coupling_period operation=\"accumulate\">10"
                       "</coupling_period>\n"
            "					<source_timelag>0</source_timelag>\n"
            "					<target_timelag>0</target_timelag>\n"
            "				</timestep>\n"
            "				<mapping_on_source>%s</mapping_on_source>\n"
            "				<interpolation_requirements\n>\n"
            "					%s\n"
            "				</interpolation_requirements>\n"
            "				<enforce_write_weight_file filename=\"%s\">%s"
                     "</enforce_write_weight_file>\n"
            "			</transient_couple>\n",
            src_comp_id * num_interpolations + k + 1, // transient_id
            l, // source component reference (1 or 2)
            src_comp_id * num_interpolations + k + 1, // source transient grid reference
            src_comp_id * num_interpolations + k + 1, // target transient grid reference
            "true", // weights are to be applied at the source component
            interpolation_xml_string[k], // interpolation configuration string
            weight_filename, // weight filename
            (write_weights)?"true":"false"
            );
          size_t temp = src_comp_id;
          src_comp_id = tgt_comp_id;
          tgt_comp_id = temp;
        }
      }
      fprintf(xmlfile, "		</couple>\n");
    }
  }
  fputs("	</couples>\n", xmlfile);

  fputs("</coupling>\n", xmlfile);

  fclose(xmlfile);
}
static int * read_mask(
  char const * filename, char const * grid_name, int const valid_mask_value) {

  size_t grid_name_len = strlen(grid_name) + 1;
  char x_dim_name[2 + grid_name_len];
  char y_dim_name[2 + grid_name_len];
  char mask_var_name[4 + grid_name_len];

  snprintf(x_dim_name, 2 + grid_name_len, "x_%s", grid_name);
  snprintf(y_dim_name, 2 + grid_name_len, "y_%s", grid_name);
  snprintf(mask_var_name, 4 + grid_name_len, "%s.msk", grid_name);

  if (!yac_file_exists(filename)) return NULL;

  int ncid, status;
  handle_error(nc_open(filename, 0, &ncid));

  int mask_var_id;
  status = nc_inq_varid(ncid, mask_var_name, &mask_var_id);

  if (status == NC_ENOTVAR) {
    handle_error(nc_close(ncid));
    return NULL;
  } else if (status != NC_NOERR)
    handle_error(status);

  int x_dim_id;
  int y_dim_id;
  handle_error(nc_inq_dimid(ncid, x_dim_name, &x_dim_id));
  handle_error(nc_inq_dimid(ncid, y_dim_name, &y_dim_id));

  size_t x_dim_len;
  size_t y_dim_len;
  handle_error(nc_inq_dimlen(ncid, x_dim_id, &x_dim_len));
  handle_error(nc_inq_dimlen(ncid, y_dim_id, &y_dim_len));

  size_t num_cells = x_dim_len * y_dim_len;

  int * cell_mask = xmalloc(num_cells * sizeof(*cell_mask));
  handle_error(nc_get_var_int(ncid, mask_var_id, cell_mask));
  for (size_t i = 0; i < num_cells; ++i)
    cell_mask[i] = cell_mask[i] == valid_mask_value;


  handle_error(nc_close(ncid));

  return cell_mask;
}

#else
#include <stdlib.h>
#include <stdio.h>
int main () {
  printf ("Examples requires compiling with NetCDF.\n");
  return EXIT_FAILURE;
}
#endif

