package de.mpg.mpimet.yacgui.ui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.JComponent;

import de.mpg.mpimet.yacgui.model.Connection;
import de.mpg.mpimet.yacgui.model.Model.Align;

/**
 * Copyright (c) 2013 Max-Planck-Institute for Meteorology, Hamburg.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Oracle or the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * GlassPane of CouplingGui. Used for drawing the connection lines.
 */
public class ConnectionsGlassPane extends JComponent {

	private static final long serialVersionUID = -44283032161666229L;

	/** list of arrows drawn on top of GUI */
	private final List<Arrow> arrows = new ArrayList<Arrow>();
	
	/** Local reference to CouplingGui (Grandparent) for quick access */
	private CouplingGui gui;
	
	public boolean isEmpty() {
		return (arrows.isEmpty());
	}
	
	public void addConnectionArrow(Connection connection) {
		
		if (connection == null) {
			return;
		}
		
		Arrow arrow = new Arrow(connection.getSourceAlign(), connection.getSourceIndex(), connection.getTargetIndex());
		if (connection.hasInterpolations()) {
			arrow.setArrowColor(Color.GREEN);
		}
		arrows.add(arrow);
		repaint();
	}
	
	public void removeArrowForTransient(Align align, int index) {
		Iterator<Arrow> iterator = arrows.iterator();
		while (iterator.hasNext()) {
			Arrow arrow = iterator.next();
			if (arrow.destIndex == index || arrow.sourceIndex == index) {
				iterator.remove();
				break;
			}
		}
		repaint();
	}
	
	public void updateConnectionArrow(Connection connection) {
		for (Arrow arrow : arrows) {
			if (arrow.sourceIndex == connection.getSourceIndex()) {
				arrow.setArrowColor(connection.hasInterpolations() ? Color.GREEN : Color.RED);
				break;
			}
		}
		repaint();
	}
	
	public void removeAll() {
		arrows.clear();
		repaint();
	}

	@Override
	protected void paintComponent(Graphics graphics) {
		super.paintComponent(graphics);
		
		//for future quick access
		if (gui == null) {
			gui = (CouplingGui) getParent().getParent();
		}
		
        for (final Arrow a : arrows) {
        	
        	boolean sourcePointVisible = gui.getComponentPanel(a.sourceAlign).isArrowPointVisible(a.sourceIndex);
        	boolean destPointVisible = gui.getComponentPanel(a.sourceAlign.getOpposite()).isArrowPointVisible(a.destIndex);
        	
    		a.paint(graphics, sourcePointVisible, destPointVisible);
        }
	}
	
	protected class Arrow {
		
		private static final int BLADE_WIDTH = 12;
		private static final double BLADES_ANGLE = Math.PI/8;
		
		private static final int ARROW_STROKE_WIDTH = 5;
		
		protected Align sourceAlign;
		protected int sourceIndex;
		protected int destIndex;
		
		private Color color;
		
		// line
		private Point sourcePos;
		private Point destPos;
		
		// coordinates of left and right blades of arrow's head
		private Point lbPos;
		private Point rbPos;

		
		public Arrow(Align sourceAlign, int sourceIndex, int destIndex) {
			this(sourceAlign, sourceIndex, destIndex, Color.RED);
		}
		
		public Arrow(Align sourceAlign, int sourceIndex, int destIndex, Color color) {
			this.sourceAlign = sourceAlign;
	        this.sourceIndex = sourceIndex;
	        this.destIndex = destIndex;
	        this.color = color;
	    }
		
		public void paint(Graphics g, boolean sourcePointVisible, boolean destPointVisible) {
			
	    	//drawing the body
			sourcePos = gui.getComponentPanel(sourceAlign).
					getRelativeCheckBoxArrowPosition(sourceIndex, ConnectionsGlassPane.this);
			destPos = gui.getComponentPanel(sourceAlign.getOpposite()).
					getRelativeCheckBoxArrowPosition(destIndex, ConnectionsGlassPane.this);
			
			Graphics2D g2d = (Graphics2D) g;
			g2d.setStroke(new BasicStroke(ARROW_STROKE_WIDTH, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
			
			//draw body
			int ovalRadius = ARROW_STROKE_WIDTH * 2;
			g2d.setColor(color);
			
			//source point we draw anyway
			if (sourcePointVisible) {
				g.fillOval(sourcePos.x - ovalRadius/2, sourcePos.y - ovalRadius/2, ovalRadius, ovalRadius);
			}
			
			//arrow we draw if both points are visible
			if (sourcePointVisible && destPointVisible) {
		    	g.drawLine(sourcePos.x, sourcePos.y, destPos.x, destPos.y);
		    	
		    	//drawing the blades
		    	int dx = destPos.x - sourcePos.x;
		    	int dy = destPos.y - sourcePos.y;
		    	
		    	double angle = Math.atan((double)dy / (sourceAlign == Align.LEFT ? dx : -dx));
		    	
		    	rbPos = new Point(destPos);
		    	rbPos.translate((int) ((sourceAlign == Align.LEFT ? -1 : 1) * BLADE_WIDTH * Math.cos(angle - BLADES_ANGLE)),
		    			(int) (-BLADE_WIDTH * Math.sin(angle - BLADES_ANGLE)));
		    	
		    	g.drawLine(destPos.x, destPos.y, rbPos.x, rbPos.y);
		    	
		    	lbPos = new Point(destPos);
		    	lbPos.translate((int) ((sourceAlign == Align.LEFT ? -1 : 1) * BLADE_WIDTH * Math.sin(Math.PI/2 - BLADES_ANGLE - angle)),
		    			(int) (-BLADE_WIDTH * Math.cos(Math.PI/2 - BLADES_ANGLE - angle)));
		    	
		    	g.drawLine(destPos.x, destPos.y, lbPos.x, lbPos.y);
			} else if (destPointVisible) { //destination point only if we didn't draw the arrow
				g.fillOval(destPos.x - ovalRadius/2, destPos.y - ovalRadius/2, ovalRadius, ovalRadius);
			}
	    }
		
		public void setArrowColor(Color color) {
			this.color = color;
		}
	}
}
