/**
 * @file config_xml.c
 *
 * @copyright Copyright  (C)  2021 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Teresa Holfeld <teresa.holfeld@zmaw.de>
 *                                 Moritz Hanke <hanke@dkrz.de>
 *
 * @author Rene Redler <rene.redler@mpimet.mpg.de>
 *         Teresa Holfeld <teresa.holfeld@zmaw.de>
 *         Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Rene Redler <rene.redler@mpimet.mpg.de>
 *             Teresa Holfeld <teresa.holfeld@zmaw.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xmlschemas.h>

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <float.h>
#include "utils.h"
#include "config_xml.h"
#include "interp_method_avg.h"
#include "interp_method_conserv.h"
#include "interp_method_creep.h"
#include "interp_method_spmap.h"

// first \<transient_couple\> is element of index 3 in \<couple\>
#define FIRSTTRANSIENTCOUPLE 2

static const char NUMERIC_FORMAT_STRING[] =
  "ERROR(%s): Value '%s' is not a numeric value.";

static int is_numeric(const char * s) {
   if (s == NULL || *s == '\0' || isspace(*s))
      return 0;
   char * p;
   strtod(s, &p);
   return (*p == '\0');
}

static int get_node_attribute_int(
  xmlNodePtr node, char const * attribute_name, int line) {

  char const * attribute_value =
    (char const *)xmlGetProp(node, (xmlChar const *)attribute_name);

  YAC_ASSERT_F(
    is_numeric(attribute_value), NUMERIC_FORMAT_STRING,
    "get_node_attribute_int", attribute_value);
  int attribute_value_int = atoi(attribute_value);
  xmlFree((void*)attribute_value);
  return attribute_value_int;
}

static char const * get_node_attribute_str(
  xmlNodePtr node, char const * attribute_name, int line) {

  char const * attribute_value =
    (char const *)xmlGetProp(node, (xmlChar const *)attribute_name);

  YAC_ASSERT_F(
    attribute_value != NULL,
    "ERROR attribute \"%s\" not found\n", attribute_name)

  char const * result = strdup((char const *)attribute_value);
  xmlFree((void*)attribute_value);
  return result;
}

static size_t get_number_of_elements_of_name(
  xmlNodePtr parent_node, const char * name) {

  xmlNodePtr child = parent_node->children;
  int n = 0;
  while (child) {
    if (!xmlStrcmp(child->name, (const xmlChar *) name)) ++n;
    child = child->next;
  }
  return (size_t)n;
}

static xmlNodePtr get_node_by_name(
  xmlNodePtr parent_node, const char * node_name) {

  xmlNodePtr node = parent_node->children;
  while (node) {
    if (!xmlStrcmp(node->name, (const xmlChar *) node_name)) return node;
    node = node->next;
  }
  return NULL;
}

static char const * dup_node_content(xmlNodePtr node, char const * node_name) {

  YAC_ASSERT_F(
    node != NULL, "ERROR xml node \"%s\" is not available\n", node_name)

  if (node->children->content == NULL) return NULL;

  return strdup((char const*)node->children->content);
}

static const char * get_date(xmlNodePtr dates_node, char const * node_name) {

  char const * date_str =
    dup_node_content(
      get_node_by_name(dates_node, node_name), node_name);

  if (date_str == NULL) return NULL;

  YAC_ASSERT_F(
    strlen(date_str) < MAX_CALENDAR_STR_LEN,
    "ERROR \"%s\" string is too long", node_name)

  return date_str;
}

static const char * get_start_date(xmlNodePtr dates_node) {

  return get_date(dates_node, "start_date");
}

static const char * get_end_date(xmlNodePtr dates_node) {

  return get_date(dates_node, "end_date");
}

static int is_valid(const xmlDocPtr doc, const char * schema_filename) {

  xmlDocPtr schema_doc = xmlReadFile(schema_filename, NULL, XML_PARSE_NONET);
  YAC_ASSERT(
    schema_doc != NULL, "Schema cannot be loaded or is not well-formed")

  xmlSchemaParserCtxtPtr parser_ctxt = xmlSchemaNewDocParserCtxt(schema_doc);
  YAC_ASSERT(
    parser_ctxt != NULL, "Unable to create parser context for schema")

  xmlSchemaPtr schema = xmlSchemaParse(parser_ctxt);
  YAC_ASSERT(schema != NULL, "Schema is not valid")

  xmlSchemaValidCtxtPtr valid_ctxt = xmlSchemaNewValidCtxt(schema);
  YAC_ASSERT(
    valid_ctxt != NULL, "Unable to create validation context for schema")

  int is_valid = xmlSchemaValidateDoc(valid_ctxt, doc) == 0;

  xmlSchemaFreeValidCtxt(valid_ctxt);
  xmlSchemaFree(schema);
  xmlSchemaFreeParserCtxt(parser_ctxt);
  xmlFreeDoc(schema_doc);

  return is_valid;
}

static enum yac_time_unit_type get_timestep_unit(xmlNodePtr coupling) {

  xmlNodePtr timestep_unit_node =
    get_node_by_name(coupling, "timestep_unit");
  char * timestep_unit_str = (char *) timestep_unit_node->children->content;

  int timestep_unit = INT_MAX;

   if      (!strcmp(timestep_unit_str, "millisecond")) timestep_unit = C_MILLISECOND;
   else if (!strcmp(timestep_unit_str, "second"))      timestep_unit = C_SECOND;
   else if (!strcmp(timestep_unit_str, "minute"))      timestep_unit = C_MINUTE;
   else if (!strcmp(timestep_unit_str, "hour"))        timestep_unit = C_HOUR;
   else if (!strcmp(timestep_unit_str, "day"))         timestep_unit = C_DAY;
   else if (!strcmp(timestep_unit_str, "month"))       timestep_unit = C_MONTH;
   else if (!strcmp(timestep_unit_str, "year"))        timestep_unit = C_YEAR;
   else if (!strcmp(timestep_unit_str, "ISO_format"))  timestep_unit = C_ISO_FORMAT;

  YAC_ASSERT_F(
    timestep_unit != INT_MAX, "Unsupported time unit '%s'.", timestep_unit_str)
  return (enum yac_time_unit_type)timestep_unit;
}

static void get_grids(
  xmlNodePtr grids_node, struct yac_couple_config_grid * grids) {

  xmlNodePtr children = grids_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"grid")) {

      xmlNodePtr grid_node = children;
      grids->name = get_node_attribute_str(grid_node, "alias_name", __LINE__);
      grids->id = get_node_attribute_int(grid_node, "id", __LINE__);

      YAC_ASSERT(
        strlen(grids->name) < MAX_FILE_NAME_LENGTH,
        "ERROR grid name is too long")

      ++grids;
    }
    children = children->next;
  }
}

static void get_transients(
  const xmlNodePtr transients_node,
  struct yac_couple_config_transient * transients) {

  xmlNodePtr children = transients_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"transient")) {

      xmlNodePtr transient_node = children;
      transients->name =
        get_node_attribute_str(
          transient_node, "transient_standard_name", __LINE__);
      transients->id = get_node_attribute_int(transient_node, "id", __LINE__);
      transients->collection_size = SIZE_MAX;
      ++transients;
    }
    children = children->next;
  }
}

static size_t get_transient_idx(
  size_t num_transients, struct yac_couple_config_transient * transients,
  int transient_id) {

  for (size_t transient_idx = 0; transient_idx < num_transients;
       ++transient_idx)
    if (transients[transient_idx].id == transient_id) return transient_idx;
  return SIZE_MAX;
}

static size_t get_grid_idx(
  size_t num_grids, struct yac_couple_config_grid * grids,
  int grid_id) {

  for (size_t grid_idx = 0; grid_idx < num_grids; ++grid_idx)
    if (grids[grid_idx].id == grid_id) return grid_idx;
  return SIZE_MAX;
}

static void get_component_transients(
  xmlNodePtr transient_grid_refs_node,
  size_t num_transients, struct yac_couple_config_transient * transients,
  size_t num_grids, struct yac_couple_config_grid * grids,
  int * transient_grid_ref_ids,
  struct yac_couple_config_component_transient * component_transients) {

  xmlNodePtr children = transient_grid_refs_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"transient_grid_ref")) {

      xmlNodePtr transient_grid_ref_node = children;
      int id =
        get_node_attribute_int(
          transient_grid_ref_node, "id", __LINE__);
      int transient_ref =
        get_node_attribute_int(
          transient_grid_ref_node, "transient_ref", __LINE__);
      int grid_ref =
        get_node_attribute_int(
          transient_grid_ref_node, "grid_ref", __LINE__);
      int collection_size =
        (size_t)get_node_attribute_int(
          transient_grid_ref_node, "collection_size", __LINE__);

      size_t transient_idx =
        get_transient_idx(num_transients, transients, transient_ref);
      YAC_ASSERT(
        transient_idx != SIZE_MAX,
        "ERROR could not find transient id from transient grid ref in "
        "list of transients")

      size_t grid_idx = get_grid_idx(num_grids, grids, grid_ref);
      YAC_ASSERT(
        grid_idx != SIZE_MAX,
        "ERROR could not find grid id from grid grid ref in list of grids")

      if (transients[transient_idx].collection_size == SIZE_MAX)
        transients[transient_idx].collection_size = collection_size;
      else
        YAC_ASSERT(
          transients[transient_idx].collection_size == collection_size,
          "ERROR(get_component_transients): collection size missmatch")

      *transient_grid_ref_ids = id;
      component_transients->transient_idx = transient_idx;
      component_transients->grid_idx = grid_idx;
      component_transients++;
      transient_grid_ref_ids++;
    }
    children = children->next;
  }
}

static void get_components(
  xmlNodePtr components_node,
  size_t num_transients, struct yac_couple_config_transient * transients,
  size_t num_grids, struct yac_couple_config_grid * grids,
  int ** transient_grid_ref_ids,
  struct yac_couple_config_component * components) {

  xmlNodePtr children = components_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"component")) {

      xmlNodePtr component_node = children;
      xmlNodePtr transient_grid_refs_node =
        get_node_by_name(component_node, "transient_grid_refs");
      size_t num_component_transients =
        get_number_of_elements_of_name(
          transient_grid_refs_node, "transient_grid_ref");
      *transient_grid_ref_ids =
        xmalloc(num_component_transients * sizeof(**transient_grid_ref_ids));
      struct yac_couple_config_component_transient * component_transients =
        xmalloc(num_component_transients * sizeof(*component_transients));
      get_component_transients(
        transient_grid_refs_node, num_transients, transients,
        num_grids, grids, *transient_grid_ref_ids, component_transients);

      xmlNodePtr name_node = get_node_by_name(component_node, "name");

      components->component_transients = component_transients;
      components->num_component_transients = num_component_transients;
      components->name = strdup((char const*)name_node->children->content);
      components->id =
        get_node_attribute_int(component_node, "id", __LINE__);
      ++transient_grid_ref_ids;
      ++components;
    }
    children = children->next;
  }
}

static enum yac_reduction_type get_reduction_type(
  xmlNodePtr coupling_period_node) {

  char * operation_str =
    (char *)xmlGetProp(coupling_period_node, (const xmlChar *)"operation");
  enum yac_reduction_type reduction_type;

  if      (!strcmp(operation_str, "accumulate")) reduction_type = ACCUMULATE;
  else if (!strcmp(operation_str, "average"))    reduction_type = TIME_AVERAGE;
  else                                           reduction_type = INSTANT;
  xmlFree((void*)operation_str);
  return reduction_type;
}

static size_t get_component_transient_idx(
  struct yac_couple_config_component * component, size_t transient_idx) {

  size_t num_component_transients = component->num_component_transients;
  struct yac_couple_config_component_transient * component_transients =
    component->component_transients;
  for (size_t component_transient_idx = 0;
       component_transient_idx < num_component_transients;
       ++component_transient_idx)
    if (component_transients[component_transient_idx].transient_idx ==
        transient_idx) return component_transient_idx;
  return SIZE_MAX;
}

static enum yac_interpolation_list get_interpolation_method(
  xmlNodePtr interpolation_node) {

  char const * method_value =
    (char const *)xmlGetProp(interpolation_node, (const xmlChar *)"method");
  enum yac_interpolation_list interpolation_method;

  if (strcmp(method_value, "average") == 0) {
     interpolation_method = AVERAGE;
  } else if (strcmp(method_value, "n-nearest_neighbor") == 0) {
     interpolation_method = N_NEAREST_NEIGHBOR;
  } else if (strcmp(method_value, "conservative") == 0) {
     interpolation_method = CONSERVATIVE;
  } else if (strcmp(method_value, "source_to_target_map") == 0) {
     interpolation_method = SOURCE_TO_TARGET_MAP;
  } else if (strcmp(method_value, "fixed_value") == 0) {
     interpolation_method = FIXED_VALUE;
  } else if (strcmp(method_value, "user_file") == 0) {
     interpolation_method = USER_FILE;
  } else if (strcmp(method_value, "check") == 0) {
     interpolation_method = CHECK;
  } else if (strcmp(method_value, "bernstein_bezier") == 0) {
    interpolation_method = BERNSTEIN_BEZIER;
  } else if (strcmp(method_value, "radial_basis_function") == 0) {
    interpolation_method = RADIAL_BASIS_FUNCTION;
  } else if (strcmp(method_value, "creep") == 0) {
    interpolation_method = CREEP;
  } else {
    interpolation_method = UNDEFINED;
  }
  xmlFree((void*)method_value);

  YAC_ASSERT(
    interpolation_method != UNDEFINED,
    "ERROR(get_interpolation_method): invalid interpolation method string")
  return interpolation_method;
}

static const char * get_interpolation_parameter_value_string(
  xmlNodePtr interpolation_node, const char * parameter) {

  xmlChar * value =
    xmlGetProp(interpolation_node, (const xmlChar *) parameter);

  if (value == NULL) return NULL;
  return (const char *) value;
}

static int get_interpolation_parameter_value_int(
  xmlNodePtr interpolation_node, const char * parameter, int default_value) {

  const char * value =
    (const char *)xmlGetProp(interpolation_node, (const xmlChar *)parameter);

  if (value == NULL) return default_value;

  YAC_ASSERT_F(
    is_numeric(value), NUMERIC_FORMAT_STRING,
    "get_interpolation_parameter_value_int", value);
  int result = atoi(value);
  xmlFree((void*)value);
  return result;
}

static double get_interpolation_parameter_value_double(
  xmlNodePtr interpolation_node, const char * parameter, double default_value) {

  const char * value =
    (const char *)xmlGetProp(interpolation_node, (const xmlChar *)parameter);

  if (value == NULL) return default_value;

  YAC_ASSERT_F(
    is_numeric(value), NUMERIC_FORMAT_STRING,
    "get_interpolation_parameter_value_double", value)
  double result = atof(value);
  xmlFree((void*)value);
  return result;
}

static int bool_str_2_int(const char * bool_str) {

  int ret = !strcmp(bool_str, "true") ||
            !strcmp(bool_str, "True") ||
            !strcmp(bool_str, "TRUE");

  YAC_ASSERT(
    ret || !strcmp(bool_str, "false") ||
           !strcmp(bool_str, "False") ||
           !strcmp(bool_str, "FALSE"),
    "ERROR(bool_str_2_int): invalid value boolean string")

  return ret;
}

static int get_interpolation_parameter_value_bool(
  xmlNodePtr interpolation_node, const char * parameter, int default_value) {

  const char * value =
    (const char *)xmlGetProp(interpolation_node, (const xmlChar *)parameter);

  if (value == NULL) return default_value;

  int result = bool_str_2_int(value);
  xmlFree((void*)value);
  return result;
}

static struct yac_interp_stack_config get_interp_stack(
  xmlNodePtr interpolation_requirements_node,
  char const * src_grid_name, char const * tgt_grid_name) {

  struct yac_interp_stack_config interp_stack;

  interp_stack.size =
    get_number_of_elements_of_name(
      interpolation_requirements_node, "interpolation");
  union yac_interp_stack_config_entry * config =
    ((interp_stack.config =
        xcalloc(interp_stack.size, sizeof(*(interp_stack.config)))));

  xmlNodePtr children = interpolation_requirements_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"interpolation")) {

      xmlNodePtr interpolation_node = children;
      config->general.type = get_interpolation_method(interpolation_node);

      YAC_ASSERT(
        config->general.type != UNDEFINED,
        "ERROR undefined interpolation method in interpolation")

      switch ((int)(config->general.type)) {

        case (AVERAGE): {

          int partial_cov =
            get_interpolation_parameter_value_bool(
              interpolation_node, "partial_coverage", 0);

          enum yac_interp_avg_weight_type reduction_type = AVG;
          const char * weighted_str =
            get_interpolation_parameter_value_string(
              interpolation_node, "weighted");

          if (weighted_str != NULL) {

            if (!strcmp(weighted_str, "distance_weighted") ||
                !strcmp(weighted_str, "DISTANCE_WEIGHTED")) {

              reduction_type = DIST;

            } else if (!strcmp(weighted_str, "arithmetic_average") ||
                       !strcmp(weighted_str, "ARITHMETIC_AVERAGE")) {

              reduction_type = AVG;

            } else if (!strcmp(weighted_str, "barycentric_coordinate") ||
                       !strcmp(weighted_str, "BARYCENTRIC_COORDINATE")) {

              reduction_type = BARY;
            } else {

              YAC_ABORT(
                "ERROR(get_interp_stack): "
                "invalid \"weighted\" value for n-nearest_neighbor")
            }

            xmlFree((void*)weighted_str);
          }

          config->average.partial_coverage = partial_cov;
          config->average.reduction_type = reduction_type;
          break;
        }
        case (N_NEAREST_NEIGHBOR): {

          int nn =
            get_interpolation_parameter_value_int(
              interpolation_node, "n", -1);

          YAC_ASSERT(
            nn != -1,
            "ERROR(get_interp_stack): N_NEAREST_NEIGHBOR requires valid n!")

          enum yac_interp_nnn_weight_type reduction_type = NNN_AVG;
          const char * weighted_str =
            get_interpolation_parameter_value_string(
              interpolation_node, "weighted");

          if (weighted_str != NULL) {

            if (!strcmp(weighted_str, "distance_weighted") ||
                !strcmp(weighted_str, "DISTANCE_WEIGHTED")) {

              reduction_type = NNN_DIST;

            } else if (!strcmp(weighted_str, "gauss_weighted") ||
                       !strcmp(weighted_str, "GAUSS_WEIGHTED")) {

              reduction_type = NNN_GAUSS;

            } else if (!strcmp(weighted_str, "arithmetic_average") ||
                       !strcmp(weighted_str, "ARITHMETIC_AVERAGE")) {

              reduction_type = NNN_AVG;

            } else {

              YAC_ABORT(
                "ERROR(get_interp_stack): "
                "invalid \"weighted\" value for n-nearest_neighbor")
            }

            xmlFree((void*)weighted_str);
          }

          double gauss_scale = -1;

          if (reduction_type == NNN_GAUSS)
            gauss_scale =
              get_interpolation_parameter_value_double(
                interpolation_node, "gauss_scale", YAC_DEFAULT_GAUSS_SCALE);

          config->n_nearest_neighbor.config.data.gauss_scale = gauss_scale;
          config->n_nearest_neighbor.config.n = (size_t)nn;
          config->n_nearest_neighbor.config.type = reduction_type;
          break;
        }
        case (CONSERVATIVE): {
          int enforced_conserv =
            get_interpolation_parameter_value_bool(
              interpolation_node, "enforced_conservation", 0);
          int partial_coverage =
            get_interpolation_parameter_value_bool(
              interpolation_node, "partial_coverage", 0);

          int order =
            get_interpolation_parameter_value_int(
              interpolation_node, "order", 1);

          YAC_ASSERT(
            (order == 1) || (order == 2),
            "ERROR(get_interp_stack): invalid value for conservative "
            "remapping XML parameter order")

          enum yac_interp_method_conserv_normalisation
            normalisation = CONSERV_DESTAREA;
          const char * normalisation_str =
            get_interpolation_parameter_value_string(
              interpolation_node, "normalisation");

          if (normalisation_str != NULL) {

            if (!strcmp(normalisation_str, "fracarea") ||
                !strcmp(normalisation_str, "FRACAREA")) {

              normalisation = CONSERV_FRACAREA;

            } else if (!(!strcmp(normalisation_str, "destarea") ||
                         !strcmp(normalisation_str, "DESTAREA"))) {

              normalisation = CONSERV_DESTAREA;
            }

            xmlFree((void*)normalisation_str);
          }

          config->conservative.order = order;
          config->conservative.enforced_conserv = enforced_conserv;
          config->conservative.partial_coverage = partial_coverage;
          config->conservative.normalisation = normalisation;
          break;
        }
        case (SOURCE_TO_TARGET_MAP):
        {

          double spread_distance =
            get_interpolation_parameter_value_double(
              interpolation_node, "spread_distance",
              YAC_DEFAULT_SPREAD_DISTANCE);

          YAC_ASSERT(
            (spread_distance >= 0.0) && (spread_distance <= 90.0),
            "ERROR(get_interp_stack): invalid spread_distance "
            "(has to be >= 0 and <= 90)")

          spread_distance *= YAC_RAD;

          enum yac_interp_spmap_weight_type weight_type = SPMAP_AVG;
          const char * weighted_str =
            get_interpolation_parameter_value_string(
              interpolation_node, "weighted");

          if (weighted_str != NULL) {

            if (!strcmp(weighted_str, "distance_weighted") ||
                !strcmp(weighted_str, "DISTANCE_WEIGHTED")) {

              weight_type = SPMAP_DIST;

            } else if (!strcmp(weighted_str, "arithmetic_average") ||
                       !strcmp(weighted_str, "ARITHMETIC_AVERAGE")) {

              weight_type = SPMAP_AVG;

            } else {

              YAC_ABORT(
                "ERROR(get_interp_stack): "
                "invalid \"weighted\" value for Soure-Point-Mapping")
            }

            xmlFree((void*)weighted_str);
          }

          config->spmap.spread_distance = spread_distance;
          config->spmap.weight_type = weight_type;
          break;
        }
        case (FIXED_VALUE):
        {
          double fixed_value =
            get_interpolation_parameter_value_double(
              interpolation_node, "user_value", DBL_MAX);

          YAC_ASSERT(
            fixed_value != DBL_MAX,
            "ERROR(get_interp_stack): start_search: FIXED_VALUE "
            "requires a value defined by the user!")

          config->fixed.value = fixed_value;
          break;
        }
        case (USER_FILE):
        {

          // get the filename containing the weights

          char const * filename =
            get_interpolation_parameter_value_string(
              interpolation_node, "filename" );
          YAC_ASSERT(
            filename != NULL,
            "ERROR(get_interp_stack): "
            "USER_FILE requires a filename containing "
            "weights defined by the user!")

          YAC_ASSERT(
            strlen(filename) < MAX_FILE_NAME_LENGTH,
            "ERROR(get_interp_stack): USER_FILE filename is too long!")
          strcpy(config->user_file.filename, filename);

          xmlFree((void*)filename);

          config->user_file.io_proc_step = -1;
          strcpy(config->user_file.src_grid_name, src_grid_name);
          strcpy(config->user_file.tgt_grid_name, tgt_grid_name);

          break;
        }
        case (CHECK):
        {
          char const * constructor_key =
            get_interpolation_parameter_value_string(
              interpolation_node, "constructor_key");

          if (constructor_key != NULL) {

            YAC_ASSERT(
              strlen(constructor_key) < MAX_ROUTINE_NAME_LENGTH,
              "ERROR(get_interp_stack): CHECK constructor_key is too long!")
            strcpy(config->check.constructor_key, constructor_key);

            xmlFree((void*)constructor_key);
          }

          char const * do_search_key =
            get_interpolation_parameter_value_string(
              interpolation_node, "do_search_key");

          if (do_search_key != NULL) {

            YAC_ASSERT(
              strlen(do_search_key) < MAX_ROUTINE_NAME_LENGTH,
              "ERROR(get_interp_stack): CHECK do_search_key is too long!")
            strcpy(config->check.do_search_key, do_search_key);

            xmlFree((void*)do_search_key);
          }

          break;
        }
        case (BERNSTEIN_BEZIER):
        {
          break;
        }
        case (RADIAL_BASIS_FUNCTION):
        {

          int nn =
            get_interpolation_parameter_value_int(
              interpolation_node, "n", 9);
          const char * weighted_str =
            get_interpolation_parameter_value_string(
              interpolation_node, "rbf_kernel");

          if (weighted_str != NULL) {

            YAC_ASSERT(
              !strcmp(weighted_str, "gauss_kernel") ||
              !strcmp(weighted_str, "GAUSS_KERNEL"),
              "ERROR(get_interp_stack): invalid value for RBF XML "
              "parameter rbf_kernel")

            xmlFree((void*)weighted_str);
          }

          config->radial_basis_function.config.data.rbf_scale =
              get_interpolation_parameter_value_double(
                interpolation_node, "rbf_scale", YAC_DEFAULT_RBF_SCALE);
          config->radial_basis_function.config.n = (size_t)nn;
          config->n_nearest_neighbor.config.type = NNN_RBF;
          break;
        }
        case (CREEP):
        {

          config->creep.creep_distance =
            get_interpolation_parameter_value_int(
                interpolation_node, "creep_distance",
                YAC_DEFAULT_CREEP_DISTANCE);
          break;
        }
        case (UNDEFINED):
          YAC_ABORT(
            "ERROR(get_interp_stack): unsupported or "
            "undefined interpolation in xml file!")
        default:
          YAC_ABORT(
            "ERROR(get_interp_stack): start_search: unknown interpolation!")
      }
      ++config;
    }
    children = children->next;
  }

  return interp_stack;
}

static void get_transient_couples(
  xmlNodePtr couple_node,
  size_t num_transients, struct yac_couple_config_transient * transients,
  size_t num_grids, struct yac_couple_config_grid * grids,
  size_t num_components, struct yac_couple_config_component * components,
  size_t * component_indices, int ** transient_grid_ref_ids,
  struct yac_couple_config_transient_couple * transient_couples) {

  xmlNodePtr children = couple_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"transient_couple")) {

      xmlNodePtr transient_couple_node = children;

      size_t transient_idx =
        get_transient_idx(
          num_transients, transients,
          get_node_attribute_int(
            transient_couple_node, "transient_id", __LINE__));
      YAC_ASSERT(
        transient_idx != SIZE_MAX,
        "ERROR could not find transient id from transient couple in "
        "list of transients")

      xmlNodePtr source_node =
        get_node_by_name(transient_couple_node, "source");
      int source_component_ref =
        get_node_attribute_int(source_node, "component_ref", __LINE__);
      YAC_ASSERT(
        (source_component_ref == 1) || (source_component_ref == 2),
        "ERROR invalid component_ref in transient_couple "
        "(has to be either 1 or 2)")
      size_t source_component_idx =
        component_indices[source_component_ref - 1];
      int source_transient_grid_ref =
        get_node_attribute_int(source_node, "transient_grid_ref", __LINE__);
      size_t source_component_transient_idx =
        get_component_transient_idx(
          components + source_component_idx, transient_idx);
      YAC_ASSERT(
        transient_grid_ref_ids
          [source_component_idx][source_component_transient_idx] ==
          source_transient_grid_ref,
        "ERROR missmatch in transient_grid_ref of transient_couple and "
        "transient_grid_refs")

      xmlNodePtr target_node =
        get_node_by_name(transient_couple_node, "target");
      size_t target_component_idx =
        component_indices[(source_component_ref - 1) ^ 1];
      int target_transient_grid_ref =
        get_node_attribute_int(target_node, "transient_grid_ref", __LINE__);
      size_t target_component_transient_idx =
        get_component_transient_idx(
          components + target_component_idx, transient_idx);
      YAC_ASSERT(
        transient_grid_ref_ids
          [target_component_idx][target_component_transient_idx] ==
          target_transient_grid_ref,
        "ERROR missmatch in transient_grid_ref of transient_couple and "
        "transient_grid_refs")

      xmlNodePtr timestep_node =
        get_node_by_name(transient_couple_node, "timestep");
      xmlNodePtr coupling_period_node =
        get_node_by_name(timestep_node, "coupling_period");
      char const * coupling_period =
        strdup((char *)coupling_period_node->children->content);
      enum yac_reduction_type coupling_period_operation =
        get_reduction_type(coupling_period_node);

      xmlNodePtr source_timestep_node =
        get_node_by_name(timestep_node, "source");
      char const * source_timestep =
        strdup((char *)source_timestep_node->children->content);
      xmlNodePtr source_timelag_node =
        get_node_by_name(timestep_node, "source_timelag");
      char * source_lag_value = (char*)source_timelag_node->children->content;
      YAC_ASSERT_F(
        is_numeric(source_lag_value), NUMERIC_FORMAT_STRING,
        "get_transient_couples", source_lag_value);
      int source_lag = atoi(source_lag_value);

      xmlNodePtr target_timestep_node =
        get_node_by_name(timestep_node, "target");
      char const * target_timestep =
        strdup((char *)target_timestep_node->children->content);
      xmlNodePtr target_timelag_node =
        get_node_by_name(timestep_node, "target_timelag");
      char * target_lag_value = (char*)target_timelag_node->children->content;
      YAC_ASSERT_F(
        is_numeric(target_lag_value), NUMERIC_FORMAT_STRING,
        "get_transient_couples", target_lag_value);
      int target_lag = atoi(target_lag_value);

      xmlNodePtr interpolation_requirements_node =
        get_node_by_name(transient_couple_node, "interpolation_requirements");
      struct yac_interp_stack_config interp_stack =
        get_interp_stack(
          interpolation_requirements_node,
          grids[
            components[source_component_idx].
              component_transients[source_component_transient_idx].
                grid_idx].name,
          grids[
            components[target_component_idx].
              component_transients[target_component_transient_idx].
                grid_idx].name);

      xmlNodePtr mapping_on_source_node =
        get_node_by_name(transient_couple_node, "mapping_on_source");
      int mapping_on_source =
        (mapping_on_source_node == NULL) ||
         bool_str_2_int(
           (char const *)(mapping_on_source_node->children->content));

      xmlNodePtr enforce_write_weight_file_node =
        get_node_by_name(
          transient_couple_node, "enforce_write_weight_file");
      int enforce_write_weight_file =
        (enforce_write_weight_file_node == NULL)?
          0:bool_str_2_int(
              (char const *)(
                enforce_write_weight_file_node->children->content));

      char const * weight_file_name =
        (char *)xmlGetProp(
          enforce_write_weight_file_node, (const xmlChar *)"filename");
      YAC_ASSERT(
        !enforce_write_weight_file || (weight_file_name != NULL),
        "Attribute \"filename\" not found.")

      transient_couples->source.component_idx = source_component_idx;
      transient_couples->source.component_transient_idx =
        source_component_transient_idx;
      transient_couples->source.timestep = source_timestep;
      transient_couples->source.lag = source_lag;
      transient_couples->target.component_idx = target_component_idx;
      transient_couples->target.component_transient_idx =
        target_component_transient_idx;
      transient_couples->target.timestep = target_timestep;
      transient_couples->target.lag = target_lag;
      transient_couples->mapping_on_source = mapping_on_source;
      transient_couples->interp_stack = interp_stack;
      transient_couples->coupling_period_operation = coupling_period_operation;
      transient_couples->coupling_period = coupling_period;
      transient_couples->enforce_write_weight_file = enforce_write_weight_file;
      transient_couples->weight_file_name = weight_file_name;
      ++transient_couples;
    }
    children = children->next;
  }
}

static size_t get_component_idx(
  size_t num_components, struct yac_couple_config_component * components,
  int component_id) {

  for (size_t component_idx = 0; component_idx < num_components;
       ++component_idx)
    if (components[component_idx].id == component_id) return component_idx;
  return SIZE_MAX;
}

static void get_couples(
  xmlNodePtr couples_node,
  size_t num_transients, struct yac_couple_config_transient * transients,
  size_t num_grids, struct yac_couple_config_grid * grids,
  size_t num_components, struct yac_couple_config_component * components,
  int ** transient_grid_ref_ids, struct yac_couple_config_couple * couples) {

  xmlNodePtr children = couples_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"couple")) {

      xmlNodePtr couple_node = children;

      size_t component_indices[2] = {
        get_component_idx(
          num_components, components,
          get_node_attribute_int(
            get_node_by_name(couple_node, "component1"),
            "component_id", __LINE__)),
        get_component_idx(
          num_components, components,
          get_node_attribute_int(
            get_node_by_name(couple_node, "component2"),
            "component_id", __LINE__))};

      YAC_ASSERT(
        (component_indices[0] != SIZE_MAX) &&
        (component_indices[1] != SIZE_MAX),
        "ERROR component id of couple not in list of components")

      size_t num_transient_couples =
        get_number_of_elements_of_name(couple_node, "transient_couple");
      struct yac_couple_config_transient_couple * transient_couples =
        xmalloc(num_transient_couples * sizeof(*transient_couples));
      get_transient_couples(
        couple_node, num_transients, transients, num_grids, grids,
        num_components, components, component_indices, transient_grid_ref_ids,
        transient_couples);

      couples->component_indices[0] = component_indices[0];
      couples->component_indices[1] = component_indices[1];
      couples->transient_couples = transient_couples;
      couples->num_transient_couples = num_transient_couples;
      ++couples;
    }
    children = children->next;
  };
}

static int redirect_stdout(xmlNodePtr coupling) {

  xmlNodePtr redirect_node = get_node_by_name(coupling, "redirect");

  if (redirect_node == NULL) return 0;

  char * value =
    (char *) xmlGetProp(redirect_node, (const xmlChar *)"redirect_stdout");

  int redirect = bool_str_2_int(value);

  if (redirect) {

    xmlFree(value);
    value =
      (char *) xmlGetProp(redirect_node, (const xmlChar *)"redirect_of_root");

    // increment only if redirect of all processes is requested
    redirect += !bool_str_2_int(value);
  }

  xmlFree(value);

  return redirect;
}

static calendarType get_calendar(xmlNodePtr dates_node) {

  xmlNodePtr calendar_node = get_node_by_name(dates_node, "calendar");
  char * calendar_str = (char *) calendar_node->children->content;
  if (strcmp(calendar_str, "proleptic-gregorian") == 0) {
    return PROLEPTIC_GREGORIAN;
  } else if (strcmp(calendar_str, "360d") == 0) {
    return YEAR_OF_360_DAYS;
  } else if (strcmp(calendar_str, "365d") == 0) {
    return YEAR_OF_365_DAYS;
  }

  return CALENDAR_NOT_SET;
}

struct yac_couple_config * yac_couple_config_xml_new(
  const char * xml_filename, const char * schema_filename) {

  // error if file does not exist
  YAC_ASSERT_F(
    yac_file_exists(xml_filename), "File %s could not be found.", xml_filename)

  // parse the XML file
  xmlDocPtr document = xmlReadFile(xml_filename, "UTF-8", 0);

  // error if parsing was not successful
  YAC_ASSERT_F(
    document != NULL, "Could not parse file %s", xml_filename)

  // validate XML against schema
  YAC_ASSERT(
    is_valid(document, schema_filename),
    "XML does not validate against schema.")

  // get root element
  xmlNodePtr coupling_node = xmlDocGetRootElement(document);

  // error if file was empty
  YAC_ASSERT_F(
    coupling_node != NULL, "File %s is empty.", xml_filename)

  struct yac_couple_config * couple_config = xmalloc(1 * sizeof(*couple_config));

  xmlNodePtr dates_node = get_node_by_name(coupling_node, "dates");
  couple_config->start_datetime = get_start_date(dates_node);
  couple_config->end_datetime = get_end_date(dates_node);
  couple_config->calendar = get_calendar(dates_node);
  couple_config->timestep_unit = get_timestep_unit(coupling_node);

  xmlNodePtr grids_node = get_node_by_name(coupling_node, "grids");
  size_t num_grids =
    ((couple_config->num_grids =
        get_number_of_elements_of_name(grids_node, "grid")));
  struct yac_couple_config_grid * grids =
    ((couple_config->grids = xmalloc(num_grids * sizeof(*grids))));
  get_grids(grids_node, grids);

  xmlNodePtr transients_node = get_node_by_name(coupling_node, "transients");
  size_t num_transients =
    ((couple_config->num_transients =
        get_number_of_elements_of_name(transients_node, "transient")));
  struct yac_couple_config_transient * transients =
    ((couple_config->transients =
        xmalloc(num_transients * sizeof(*transients))));
  get_transients(transients_node, transients);

  xmlNodePtr components_node = get_node_by_name(coupling_node, "components");
  size_t num_components =
    ((couple_config->num_components =
        get_number_of_elements_of_name(components_node, "component")));
  struct yac_couple_config_component * components =
    ((couple_config->components =
        xmalloc(num_components * sizeof(*components))));
  int ** transient_grid_ref_ids =
    xmalloc(num_components * sizeof(*transient_grid_ref_ids));
  get_components(
    components_node, num_transients, transients, num_grids, grids,
    transient_grid_ref_ids, components);

  xmlNodePtr couples_node = get_node_by_name(coupling_node, "couples");
  size_t num_couples =
    ((couple_config->num_couples =
        get_number_of_elements_of_name(couples_node, "couple")));
  struct yac_couple_config_couple * couples =
    ((couple_config->couples = xmalloc(num_couples * sizeof(*couples))));
  get_couples(
    couples_node, num_transients, transients, num_grids, grids,
    num_components, components, transient_grid_ref_ids, couples);

  for (size_t component_idx = 0; component_idx < num_components;
       ++component_idx)
    free(transient_grid_ref_ids[component_idx]);
  free(transient_grid_ref_ids);

  couple_config->redirect_stdout = redirect_stdout(coupling_node);

  // destroy parser
  xmlCleanupParser();
  xmlFreeDoc(document);

  return couple_config;
}
