/**
 * @file couple_config.c
 *
 * @copyright Copyright  (C)  2021 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "couple_config.h"
#include "yac_interface.h"
#include "interp_method_avg.h"
#include "interp_method_check.h"
#include "interp_method_conserv.h"
#include "interp_method_creep.h"
#include "interp_method_nnn.h"
#include "interp_method_fixed.h"
#include "interp_method_file.h"
#include "interp_method_hcsbb.h"
#include "interp_method_spmap.h"

void yac_couple_config_delete(struct yac_couple_config * couple_config) {

  free((void*)couple_config->start_datetime);
  free((void*)couple_config->end_datetime);

  for (size_t grid_idx = 0; grid_idx < couple_config->num_grids; ++grid_idx)
    free((void*)couple_config->grids[grid_idx].name);
  free(couple_config->grids);

  for (size_t component_idx = 0;
       component_idx < couple_config->num_components; ++component_idx) {
    free(couple_config->components[component_idx].component_transients);
    free((void*)couple_config->components[component_idx].name);
  }
  free(couple_config->components);

  for (size_t transient_idx = 0;
       transient_idx < couple_config->num_transients; ++transient_idx)
    free((void*)couple_config->transients[transient_idx].name);
  free(couple_config->transients);

  for (size_t couple_idx = 0; couple_idx < couple_config->num_couples;
       ++couple_idx) {
    for (size_t transient_couple_idx = 0;
         transient_couple_idx <
           couple_config->couples[couple_idx].num_transient_couples;
         ++transient_couple_idx) {

      free(
        (void*)couple_config->
          couples[couple_idx].
            transient_couples[transient_couple_idx].
              source.timestep);
      free(
        (void*)couple_config->
          couples[couple_idx].
            transient_couples[transient_couple_idx].
              target.timestep);

      free(
        couple_config->
          couples[couple_idx].
            transient_couples[transient_couple_idx].
              interp_stack.config);

      free(
        (void*)couple_config->
          couples[couple_idx].
            transient_couples[transient_couple_idx].
              coupling_period);

      free(
        (void*)couple_config->
          couples[couple_idx].
            transient_couples[transient_couple_idx].
              weight_file_name);
    }
    free(
      (void*)couple_config->
        couples[couple_idx].
          transient_couples);
  }
  free(couple_config->couples);
  free(couple_config);
}

size_t yac_couple_config_get_num_couples(
  struct yac_couple_config * couple_config) {

  return couple_config->num_couples;
}

static void check_couple_idx(
  struct yac_couple_config * couple_config, size_t couple_idx,
  char const * routine_name, int line) {

  YAC_ASSERT_F(
    (couple_idx >= 0) && (couple_idx < couple_config->num_couples),
    "ERROR(%s:%d:%s): invalid couple_idx", __FILE__, line, routine_name);
}

size_t yac_couple_config_get_num_couple_transients(
  struct yac_couple_config * couple_config, size_t couple_idx) {

  check_couple_idx(
    couple_config, couple_idx, "yac_couple_config_get_num_couple_transients",
    __LINE__);

  return couple_config->couples[couple_idx].num_transient_couples;
}

void yac_couple_config_get_couple_component_ids(
  struct yac_couple_config * couple_config, size_t couple_idx,
  int couple_component_ids[2]) {

  check_couple_idx(
    couple_config, couple_idx, "yac_couple_config_get_couple_component_ids",
    __LINE__);

  for (int i = 0; i < 2; ++i)
    couple_component_ids[i] =
      couple_config->components[
        couple_config->couples[couple_idx].component_indices[i]].id;
}


int yac_couple_config_component_name_is_valid(
  struct yac_couple_config * couple_config, char const * component_name) {

  YAC_ASSERT(
    strlen(component_name) <= YAC_MAX_CHARLEN,
    "ERROR(yac_couple_config_component_name_is_valid): "
    "component name is too long (maximum is YAC_MAX_CHARLEN)")

  for (size_t component_idx = 0; component_idx < couple_config->num_components;
       ++component_idx)
    if (!strcmp(component_name, couple_config->components[component_idx].name))
      return 1;
  return 0;
}

size_t yac_couple_config_get_num_components(
  struct yac_couple_config * couple_config) {

  return couple_config->num_components;
}

static void check_component_idx(
  struct yac_couple_config * couple_config, size_t component_idx,
  char const * routine_name, int line) {

  YAC_ASSERT_F(
    (component_idx >= 0) && (component_idx < couple_config->num_components),
    "ERROR(%s:%d:%s): invalid component_idx", __FILE__, line, routine_name)
}

int yac_couple_config_get_component_id(
  struct yac_couple_config * couple_config, size_t component_idx) {

  check_component_idx(
    couple_config, component_idx,
    "yac_couple_config_get_component_id", __LINE__);

  return couple_config->components[component_idx].id;
}

char const * yac_couple_config_get_component_name(
  struct yac_couple_config * couple_config, size_t component_idx) {

  check_component_idx(
    couple_config, component_idx,
    "yac_couple_config_get_component_name", __LINE__);

  return couple_config->components[component_idx].name;
}

char const * yac_couple_config_get_component_name_by_id(
  struct yac_couple_config * couple_config, int component_id) {

  char const * comp_name = NULL;
  for (size_t component_idx = 0;
       (component_idx < couple_config->num_components) && (comp_name == NULL);
       ++component_idx)
    if (couple_config->components[component_idx].id == component_id)
      comp_name = couple_config->components[component_idx].name;

  YAC_ASSERT_F(
    comp_name != NULL,
    "ERROR(yac_couple_config_get_component_name_by_id): "
    "No component with ID %d found.", component_id)
  return comp_name;
}

int yac_couple_config_get_component_id_by_name(
  struct yac_couple_config * couple_config, char const * component_name) {

  int comp_id = INT_MAX;
  for (size_t component_idx = 0;
       (component_idx < couple_config->num_components) && (comp_id == INT_MAX);
       ++component_idx)
    if (!strcmp(couple_config->components[component_idx].name, component_name))
      comp_id = couple_config->components[component_idx].id;

  YAC_ASSERT_F(
    comp_id != INT_MAX,
    "ERROR(yac_couple_config_get_component_id_by_name): "
    "No component with name \"%s\" found.", component_name)
  return comp_id;
}

static void check_transient_couple_idx(
  struct yac_couple_config * couple_config, size_t couple_idx,
  size_t transient_couple_idx, char const * routine_name, int line) {

  check_couple_idx(couple_config, couple_idx, routine_name, __LINE__);

  YAC_ASSERT_F(
    (transient_couple_idx >= 0) &&
    (transient_couple_idx <
       couple_config->couples[couple_idx].num_transient_couples),
    "ERROR(%s:%d:%s): invalid transient_couple_idx",
    __FILE__, line, routine_name)
}

struct yac_interp_stack_config yac_couple_config_get_interp_stack(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_interp_stack", __LINE__);

  return
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].interp_stack;
}

void yac_couple_config_get_transient_grid_names(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx,
  char const ** src_grid_name, char const ** tgt_grid_name) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_transient_grid_names", __LINE__);

  size_t src_component_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].
        source.component_idx;
  size_t src_component_transient_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].
        source.component_transient_idx;

  size_t tgt_component_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].
        target.component_idx;
  size_t tgt_component_transient_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].
        target.component_transient_idx;

  *src_grid_name =
    couple_config->grids[
      couple_config->components[src_component_idx].
        component_transients[src_component_transient_idx].grid_idx].name;
  *tgt_grid_name =
    couple_config->grids[
      couple_config->components[tgt_component_idx].
        component_transients[tgt_component_transient_idx].grid_idx].name;
}


size_t yac_couple_config_get_transient_collection_size(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_transient_collection_size", __LINE__);

  size_t component_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].source.component_idx;
  size_t component_transient_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].source.component_transient_idx;

  return
    couple_config->transients[
      couple_config->components[component_idx].
        component_transients[component_transient_idx].
          transient_idx].collection_size;
}

void yac_couple_config_get_transient_couple_component_ids(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx,
  int * src_component_id, int * tgt_component_id) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_transient_couple_component_ids", __LINE__);

  *src_component_id =
    couple_config->components[
      couple_config->couples[couple_idx].
        transient_couples[transient_couple_idx].source.component_idx].id;
  *tgt_component_id =
    couple_config->components[
      couple_config->couples[couple_idx].
        transient_couples[transient_couple_idx].target.component_idx].id;
}

char const * yac_couple_config_get_transient_name(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_transient_name", __LINE__);

  size_t component_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].source.component_idx;
  size_t component_transient_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].source.component_transient_idx;

  return
    couple_config->transients[
      couple_config->components[component_idx].
        component_transients[component_transient_idx].
          transient_idx].name;
}

int yac_couple_config_mapping_on_source(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_mapping_on_source", __LINE__);

  return
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          mapping_on_source;
}

enum yac_time_unit_type yac_couple_config_get_timestep_unit(
  struct yac_couple_config * couple_config) {

  return couple_config->timestep_unit;
}

char const * yac_couple_config_get_coupling_period(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_coupling_period", __LINE__);

  return
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          coupling_period;
}

void yac_couple_config_get_source_timestep(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx,
  char const ** timestep, int * lag) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_source_timestep", __LINE__);

  *timestep =
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          source.timestep;
  *lag =
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          source.lag;
}

void yac_couple_config_get_target_timestep(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx,
  char const ** timestep, int * lag) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_target_timestep", __LINE__);

  *timestep =
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          target.timestep;
  *lag =
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          target.lag;
}

enum yac_reduction_type yac_couple_config_get_coupling_period_operation(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_coupling_period_operation", __LINE__);

  return
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          coupling_period_operation;
}

void yac_couple_config_set_datetime(
  struct yac_couple_config * couple_config,
  char const * start, char const * end) {

  if ((start != NULL) || (strlen(start) > 0)) {

    free((void*)couple_config->start_datetime);
    couple_config->start_datetime = strdup(start);
  }

  if ((end != NULL) || (strlen(end) > 0)) {

    free((void*)couple_config->end_datetime);
    couple_config->end_datetime = strdup(end);
  }
}

char const * yac_couple_config_get_start_datetime(
  struct yac_couple_config * couple_config) {

  return couple_config->start_datetime;
}

char const * yac_couple_config_get_end_datetime(
  struct yac_couple_config * couple_config) {

  return couple_config->end_datetime;
}

int yac_couple_config_get_grid_id_by_name(
  struct yac_couple_config * couple_config, char const * grid_name) {

  int grid_id = INT_MAX;
  for (size_t grid_idx = 0;
      (grid_idx < couple_config->num_grids) && (grid_id == INT_MAX); ++grid_idx)
    if (!strcmp(couple_config->grids[grid_idx].name, grid_name))
      grid_id = couple_config->grids[grid_idx].id;

  YAC_ASSERT_F(
    grid_id != INT_MAX,
    "ERROR(yac_couple_config_get_grid_id_by_name): "
    "No grid with name \"%s\" found.", grid_name)
  return grid_id;
}

int yac_couple_config_enforce_write_weight_file(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_enforce_write_weight_file", __LINE__);

  return
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          enforce_write_weight_file;
}

char const * yac_couple_config_get_weight_file_name(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_weight_file_name", __LINE__);

  static char dummy[] = "\0";
  char const * weight_file_name =
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          weight_file_name;

  return (weight_file_name != NULL)?weight_file_name:dummy;
}

int yac_couple_config_redirect_stdout(
  struct yac_couple_config * couple_config) {

  return couple_config->redirect_stdout;
}

int yac_couple_config_contains_grid_name(
  struct yac_couple_config * couple_config, char const * grid_name) {

  for (size_t grid_idx = 0; grid_idx < couple_config->num_grids;
       ++grid_idx)
    if (!strcmp(couple_config->grids[grid_idx].name, grid_name))
      return 1;
  return 0;
}

int yac_couple_config_get_transient_id_by_name(
  struct yac_couple_config * couple_config, char const * transient_name) {

  int transient_id = INT_MAX;
  for (size_t transient_idx = 0;
       (transient_idx < couple_config->num_transients) &&
       (transient_id == INT_MAX); ++transient_idx)
    if (!strcmp(couple_config->transients[transient_idx].name, transient_name))
      transient_id = couple_config->transients[transient_idx].id;

  YAC_ASSERT_F(
    transient_id != INT_MAX,
    "ERROR(yac_couple_config_get_transient_id_by_name): "
    "No transient with name \"%s\" found.", transient_name)
  return transient_id;
}

int yac_couple_config_transient_config_is_valid(
  struct yac_couple_config * couple_config,
  int component_id, int grid_id, int transient_id) {

  size_t component_idx;
  for (component_idx = 0; component_idx < couple_config->num_components;
       ++component_idx)
    if (couple_config->components[component_idx].id == component_id) break;
  if (component_idx >= couple_config->num_components) return 0;

  size_t grid_idx;
  for (grid_idx = 0; grid_idx < couple_config->num_grids;
       ++grid_idx)
    if (couple_config->grids[grid_idx].id == grid_id) break;
  if (grid_idx >= couple_config->num_grids) return 0;

  size_t transient_idx;
  for (transient_idx = 0; transient_idx < couple_config->num_transients;
       ++transient_idx)
    if (couple_config->transients[transient_idx].id == transient_id) break;
  if (transient_idx >= couple_config->num_transients) return 0;

  for (size_t component_transient_idx = 0;
       component_transient_idx <
         couple_config->components[component_idx].num_component_transients;
       ++component_transient_idx)
    if ((couple_config->
           components[component_idx].
             component_transients[component_transient_idx].
               transient_idx == transient_idx) &&
        (couple_config->
           components[component_idx].
             component_transients[component_transient_idx].
               grid_idx == grid_idx)) return 1;
  return 0;
}

calendarType yac_couple_config_get_calendar(
  struct yac_couple_config * couple_config) {

  return couple_config->calendar;
}

int yac_interp_stack_config_entry_compare(void const * a_, void const * b_) {

  union yac_interp_stack_config_entry const * a =
   (union yac_interp_stack_config_entry const *)a_;
  union yac_interp_stack_config_entry const * b =
   (union yac_interp_stack_config_entry const *)b_;

  if (a->general.type != b->general.type)
    return (a->general.type > b->general.type) -
           (a->general.type < b->general.type);

  switch(a->general.type) {
    case (UNDEFINED): {
      return 0;
    }
    case (AVERAGE): {
      if (a->average.reduction_type != b->average.reduction_type)
        return (a->average.reduction_type > b->average.reduction_type) -
               (a->average.reduction_type < b->average.reduction_type);
      return (a->average.partial_coverage > b->average.partial_coverage) -
             (a->average.partial_coverage < b->average.partial_coverage);
    }
    case (RADIAL_BASIS_FUNCTION):
    case (N_NEAREST_NEIGHBOR): {
      if (a->n_nearest_neighbor.config.n !=
          b->n_nearest_neighbor.config.n)
        return (a->n_nearest_neighbor.config.n >
                b->n_nearest_neighbor.config.n) -
               (a->n_nearest_neighbor.config.n <
                b->n_nearest_neighbor.config.n);
      if (a->n_nearest_neighbor.config.type !=
          b->n_nearest_neighbor.config.type)
        return (a->n_nearest_neighbor.config.type >
                b->n_nearest_neighbor.config.type) -
               (a->n_nearest_neighbor.config.type <
                b->n_nearest_neighbor.config.type);
      switch (a->n_nearest_neighbor.config.type) {
        case (NNN_GAUSS):
          return (a->n_nearest_neighbor.config.data.gauss_scale >
                  b->n_nearest_neighbor.config.data.gauss_scale) -
                 (a->n_nearest_neighbor.config.data.gauss_scale <
                  b->n_nearest_neighbor.config.data.gauss_scale);
        case (NNN_RBF):
          return (a->n_nearest_neighbor.config.data.rbf_scale >
                  b->n_nearest_neighbor.config.data.rbf_scale) -
                 (a->n_nearest_neighbor.config.data.rbf_scale <
                  b->n_nearest_neighbor.config.data.rbf_scale);
        default:
          return 0;
      };
    }
    case (CONSERVATIVE): {
      if (a->conservative.order !=
          b->conservative.order)
        return (a->conservative.order >
                b->conservative.order) -
               (a->conservative.order <
                b->conservative.order);
      if (a->conservative.enforced_conserv !=
          b->conservative.enforced_conserv)
        return (a->conservative.enforced_conserv >
                b->conservative.enforced_conserv) -
               (a->conservative.enforced_conserv <
                b->conservative.enforced_conserv);
      if (a->conservative.partial_coverage !=
          b->conservative.partial_coverage)
        return (a->conservative.partial_coverage >
                b->conservative.partial_coverage) -
               (a->conservative.partial_coverage <
                b->conservative.partial_coverage);
      return (a->conservative.normalisation >
              b->conservative.normalisation) -
              (a->conservative.normalisation <
              b->conservative.normalisation);
    }
    case (SOURCE_TO_TARGET_MAP): {
      if (a->spmap.spread_distance !=
          b->spmap.spread_distance)
        return (a->spmap.spread_distance >
                b->spmap.spread_distance) -
               (a->spmap.spread_distance <
                b->spmap.spread_distance);
      return (a->spmap.weight_type >
              b->spmap.weight_type) -
             (a->spmap.weight_type <
              b->spmap.weight_type);
    }
    case (FIXED_VALUE): {
      return (a->fixed.value >
              b->fixed.value) -
             (a->fixed.value <
              b->fixed.value);
    }
    case (USER_FILE): {
      int ret;
      if ((ret = strcmp(a->user_file.filename, b->user_file.filename)))
        return ret;
      if ((ret = strcmp(a->user_file.src_grid_name, b->user_file.src_grid_name)))
        return ret;
      if ((ret = strcmp(a->user_file.tgt_grid_name, b->user_file.tgt_grid_name)))
        return ret;
      return (a->user_file.io_proc_step >
              b->user_file.io_proc_step) -
             (a->user_file.io_proc_step <
              b->user_file.io_proc_step);
    }
    case (CHECK): {
      int ret;
      if ((ret = (a->check.constructor_key == NULL) -
                 (b->check.constructor_key == NULL))) return ret;
      if ((ret = strcmp(a->check.constructor_key, b->check.constructor_key)))
        return ret;
      if ((ret = (a->check.do_search_key == NULL) -
                 (b->check.do_search_key == NULL))) return ret;
      return strcmp(a->check.do_search_key, b->check.do_search_key);
    };
    case (BERNSTEIN_BEZIER): {
      return 0;
    };
    case (CREEP): {
      return (a->creep.creep_distance >
              b->creep.creep_distance) -
             (a->creep.creep_distance <
              b->creep.creep_distance);
    }
    default: {
      YAC_ABORT(
        "ERROR(yac_interp_stack_config_entry_compare): "
        "invalid interpolation type")
      return 0;
    }
  };
}

int yac_interp_stack_config_compare(void const * a_, void const * b_) {

  struct yac_interp_stack_config const * a =
   (struct yac_interp_stack_config const *)a_;
  struct yac_interp_stack_config const * b =
   (struct yac_interp_stack_config const *)b_;

  int ret;
  if ((ret = (a->size > b->size) - (a->size < b->size))) return ret;

  size_t stack_size = a->size;
  for (size_t method_idx = 0; method_idx < stack_size; ++method_idx)
    if ((ret =
           yac_interp_stack_config_entry_compare(
             &(a->config[method_idx]), &(b->config[method_idx]))))
      return ret;
  return 0;
}

struct interp_method ** yac_interp_stack_config_generate(
  struct yac_interp_stack_config interp_stack) {

  size_t interp_stack_size = interp_stack.size;
  struct interp_method ** method_stack =
    xmalloc((interp_stack_size + 1) * sizeof(*method_stack));
  method_stack[interp_stack_size] = NULL;

  for (size_t i = 0; i < interp_stack_size; ++i) {

    switch((int)(interp_stack.config[i].general.type)) {
      case(AVERAGE): {
        enum yac_interp_avg_weight_type weight_type =
          interp_stack.config[i].average.reduction_type;
        int partial_coverage =
          (int)interp_stack.config[i].average.partial_coverage;
        method_stack[i] =
          yac_interp_method_avg_new(weight_type, partial_coverage);
        break;
      }
      case(CONSERVATIVE): {
        int order =
          interp_stack.config[i].conservative.order;
        int enforced_conserv =
          interp_stack.config[i].conservative.enforced_conserv;
        int partial_coverage =
          interp_stack.config[i].conservative.partial_coverage;
        enum yac_interp_method_conserv_normalisation normalisation =
          interp_stack.config[i].conservative.normalisation;

        method_stack[i] =
          yac_interp_method_conserv_new(
            order, enforced_conserv, partial_coverage, normalisation);
        break;
      }
      case(FIXED_VALUE): {
        double fixed_value =
          interp_stack.config[i].fixed.value;
        method_stack[i] = yac_interp_method_fixed_new(fixed_value);
        break;
      }
      case(USER_FILE): {
        char const * weight_file_name =
          interp_stack.config[i].user_file.filename;
        int io_proc_step =
          interp_stack.config[i].user_file.io_proc_step;
        char const * src_grid_name =
          interp_stack.config[i].user_file.src_grid_name;
        char const * tgt_grid_name =
          interp_stack.config[i].user_file.tgt_grid_name;
        method_stack[i] =
          yac_interp_method_file_new(
            weight_file_name, io_proc_step, src_grid_name, tgt_grid_name);
        break;
      }
      case(CHECK): {
        func_constructor constructor_callback;
        void * constructor_user_data;
        func_do_search do_search_callback;
        void * do_search_user_data;

        yac_interp_method_check_get_constructor_callback(
          interp_stack.config[i].check.constructor_key,
          &constructor_callback, &constructor_user_data);
        yac_interp_method_check_get_do_search_callback(
          interp_stack.config[i].check.do_search_key,
          &do_search_callback, &do_search_user_data);

        method_stack[i] =
          yac_interp_method_check_new(
            constructor_callback, constructor_user_data,
            do_search_callback, do_search_user_data);
        break;
      }
      case (N_NEAREST_NEIGHBOR): {
        method_stack[i] =
          yac_interp_method_nnn_new(
            interp_stack.config[i].
              n_nearest_neighbor.config);
        break;
      }
      case (BERNSTEIN_BEZIER): {
        method_stack[i] =
          yac_interp_method_hcsbb_new();
        break;
      }
      case (RADIAL_BASIS_FUNCTION): {
        method_stack[i] =
          yac_interp_method_nnn_new(
            interp_stack.config[i].
              radial_basis_function.config);
        break;
      }
      case (SOURCE_TO_TARGET_MAP): {
        method_stack[i] =
          yac_interp_method_spmap_new(
            interp_stack.config[i].spmap.spread_distance,
            interp_stack.config[i].spmap.weight_type);
        break;
      }
      case (CREEP): {
        method_stack[i] =
          yac_interp_method_creep_new(
            interp_stack.config[i].creep.creep_distance);
        break;
      }
      default:
        YAC_ABORT(
          "ERROR(yac_interp_stack_config_generate): "
          "unsupported interpolation method")
    };
  }

  return method_stack;
}
